# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2014 Intel
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Manifest file parser.
"""

import os
from vsm.manifest import sys_info
from vsm import exception
from vsm import flags
from vsm import db
from vsm import context
from vsm.openstack.common import log as logging

LOG = logging.getLogger(__name__)
FLAGS = flags.FLAGS

class ManifestParser(object):
    """To parse manifest file

    It's notable that we can parser cluster.manifest
    and server.manifest file. By default we analysis
    server.manifest file. Use as below:

        To analysis server.manifest:
        mpser = ManifestParser()
        mpser = ManifestParser('/etc/manifest/server.manifest', True)

        To analysis cluster.manifest:
        mpser = ManifestParser('/etc/manifest/cluster.manifest', False)

    By using this parser, you hould decide which [].manifest file will be used.
    """

    def __init__(self,
                 fpath=FLAGS.server_manifest,
                 is_server_manifest=True):
        """Set the file path, and which file to parser."""

        # Define two lists below. _disk_type_list need two columns.
        # _single_type_list need just one column.

        #self._disk_type_list = ['ssd', '7200_rpm_sata', '10krpm_sas',
        #       'ssd_cached_7200rpm_sata', 'ssd_cached_10krpm_sas']

        self._context = context.get_admin_context()
        #self._disk_type_list = self._get_disk_type_list()
        self._single_type_list = ['vsm_controller_ip', 'role']

        # Sections in cluster.manifest
        self._cluster_names = ['storage_class',
                               'storage_group', 'cluster',
                               'file_system',
                               'management_addr', 'ceph_public_addr',
                               'ceph_cluster_addr',
                               'storage_group_near_full_threshold',
                               'storage_group_full_threshold',
                               'ceph_near_full_threshold',
                               #'osd_heartbeat_interval',
                               #'osd_heartbeat_grace',
                               'ceph_full_threshold'
                                ]

        self._is_server_manifest = is_server_manifest
        self._file_path = fpath
        if not fpath:
            self._file_path = FLAGS.server_manifest

        if is_server_manifest == False:
            self._file_path = fpath or FLAGS.cluster_manifest

        if not os.path.exists(self._file_path):
            sys_info.wait_disk_ready(self._file_path)
        self._lines = None
        self._read_lines()
        self._check_key_words_exists()
        self._name_dicts = {}
        self._map = {}

    @classmethod
    def parse_cluster_manifest(cls,
                           fpath=FLAGS.cluster_manifest,
                           is_server_manifest=False):
        """Class method to support cluster manifest.

        Support one parser to parse multiple files:
              1. parser = ManifestParser(filepath, False)
                 json_txt = parser.json()
                 json_txt = parser.parse_file(another_file, False)

        """
        inst = cls(fpath, is_server_manifest)
        return inst.format_to_json()

    @classmethod
    def parse_server_manifest(cls,
                   fpath=FLAGS.server_manifest,
                   is_server_manifest=True):
        """Class method to support different usage.

        Support one parser to parse multiple files:
            1. parser = ManifestParser(filepath, True)
               json_txt = parser.json()
               json_txt = parser.parse_file(another_file, True)
        """
        inst = cls(fpath, is_server_manifest)
        return inst.format_to_json()

    def _get_disk_type_list(self):
        group_list = db.storage_group_get_all(self._context)
        disk_type_list = []
        if group_list:
            for item in group_list:
               disk_type_list.append(item['storage_class']) 
        return disk_type_list

    def _read_lines(self):
        """Read self._file_path and return the lines needed.

        Need to do filter operations:
            1. lines full with spaces
            2. comments
            3. comments starts with spaces.
        """
        #print 'self._file_path==',self._file_path
        lines = open(self._file_path).readlines()
        self._lines = []
        for single_line in lines:
            #print 'line^^^^^^^^^^^^^^',single_line
            single_line = single_line.strip()
            if not single_line.startswith("#") and len(single_line):
                self._lines.append(single_line)

    def _check_single_key_word_exists(self, name):
        """Check name exists in this file."""
        # Find the right position of the key word.
        right_pos = -1
        for pos, val in enumerate(self._lines):
            if val.find(name) != -1:
                right_pos = pos
                break

        # If not find, do not raise error to support server_list.
        if right_pos == -1:
            LOG.info("[warning] can not find %s" % name)
            return len(self._lines)

        return right_pos

    def _check_key_words_exists(self):
        """Check all the key words are exists."""

        if not self._lines or len(self._lines) == 0:
            LOG.error('Have not read any lines from file')
            raise exception.NotFound('self._lines in parser are None.')

        #key_words = self._single_type_list + self._disk_type_list
        key_words = self._single_type_list
        for kword in key_words:
            self._check_single_key_word_exists(kword)

    # TODO : need to support multiple controller ip
    def _vsm_controller_ip(self):
        """Get the vsm's controller IP

        It's notable that this IP not just refer to vsm-api node.
        It can refer to two kinds of nodes which can reach to:

            1. Proxy Node with Load Banlance.
            2. VSM-API node. We will have multiple api node running
               in the futhure.
        """
        ip_list = self._get_segment('vsm_controller_ip')
        ip_list = ip_list['single']

        # If there are multiple controller's ip, raise error.
        # We just ignore if these IPs are refer to the same node.
        # We just support one IP here.
        if len(ip_list) > 1:
            LOG.error('Just support one controller IP for VSM.')
            raise exception.DuplicateControllerIPs()

        if len(ip_list) == 0:
            LOG.error('No controller IP for VSM.')
            raise exception.NotFound()

        return ip_list[0]

    #TODO: Delete it later
    #def _number_of_disks(self):
    #    """Return the number of disks."""
    #    counter = 0
    #    for name in self._disk_type_list:
    #        name_dict = self._get_segment(name)
    #        counter = counter + len(name_dict['first'])

    #    return counter

    def _get_segment_from_record_dict(self, name):
        """Get the segment from self._name_dicts if exists."""
        if self._name_dicts.get(name, None):
            return self._name_dicts[name]
        return None

    def _get_cluster_dict(self):
        """Return dict info of cluster."""
        def _get_cluster_value(ktype):
            rpos = self._check_single_key_word_exists(ktype)
            #print 'self.lines=%s-----and pos=%s----'%(self._lines,rpos)
            for pos in (rpos + 1, len(self._lines) - 1):
                single_line = self._lines[pos]
                if single_line.find("[") != -1:
                    break

                seg_list = single_line.split()
                if len(seg_list) == 1:
                    return seg_list[0]

        try:
            mkfs_option = _get_cluster_value('mkfs_option')
        except:
            mkfs_option = None

        try:
            mount_option = _get_cluster_value('mount_option')
        except:
            mount_option = None

        ret = {'cluster_name': _get_cluster_value('cluster'),
               'file_system': _get_cluster_value('file_system'),
               'mkfs_option': mkfs_option,
               'mount_option': mount_option,
               'public_addr': _get_cluster_value('management_addr'),
               'secondary_public_addr':
                              _get_cluster_value('ceph_public_addr'),
               'cluster_addr': _get_cluster_value('ceph_cluster_addr'),
               'osd_heartbeat_interval': _get_cluster_value('osd_heartbeat_interval'),
               'osd_heartbeat_grace': _get_cluster_value('osd_heartbeat_grace')}

        return ret

    def _get_segment(self, name):
        """Get segment of manifest.

        Define manifest as below:
            [ssd]
            /dev/1 /dev/2
            /dev/3 /dev/4

        Put this kind into dict as :
            {'name': 'ssd',
             'first': ['/dev/1', '/dev/3'],
             'second': ['/dev/2', '/dev/4']}

        Also have one style:
            [role]
            storage
            monitor
            mds

        Put this kind into dict as:
            {'name': 'role',
             'single': ['storage', 'monitor', 'mds']}

        Then we change this dict into json as you wanted.
        """
        if name == "cluster":
            return self._get_cluster_dict()

        ret = self._get_segment_from_record_dict(name)
        if ret:
            return ret
        # Make sure this key word exists.
        right_pos = self._check_single_key_word_exists(name)
        ret = {'name': name, 'single': [], 'first': [],
               'second': [], 'third': [], 'fourth': [],
               'fifth': []}

        # Read key word's value into dict.
        # Raise error if find un-matching columns.
        enter_single_column = False
        enter_two_columns = False
        enter_three_columns = False
        enter_four_columns = False
        find_error = False
        for pos in range(right_pos + 1, len(self._lines)):
            single_line = self._lines[pos]
            find_error = enter_single_column and enter_two_columns
            find_error = find_error and enter_three_columns
            find_error = find_error and enter_four_columns

            if not single_line.startswith("[") and find_error == False:
                columns = single_line.split()
                if len(columns) == 5:
                    enter_four_columns = True 
                    ret['first'].append(columns[0])
                    ret['second'].append(columns[1])
                    ret['third'].append(columns[2])
                    ret['fourth'].append(columns[3])
                    ret['fifth'].append(columns[4])
                if len(columns) == 3:
                    enter_three_columns = True
                    ret['first'].append(columns[0])
                    ret['second'].append(columns[1])
                    ret['third'].append(columns[2])
                if len(columns) == 2:
                    enter_two_columns = True
                    ret['first'].append(columns[0])
                    ret['second'].append(columns[1])
                if len(columns) == 1:
                    enter_single_column = True
                    # Write this to support:
                    # storage,monitor,mds
                    parts = columns[0].split(',')
                    ret['single'] = ret['single'] +\
                         [segs for segs in parts if len(segs) > 0]
            else:
                break

        if find_error:
            LOG.error('Columns not match for %s' % name)
            raise
        self._name_dicts['name'] = ret
        return ret

    def _dict_insert_controller_ip(self):
        """Insert vsm controller ip into json format dict."""
        self._map['vsm_controller_ip'] = self._vsm_controller_ip()
        self._map['cluster_ip'] = self._vsm_controller_ip()

    def _dict_insert_role(self):
        """Insert role into json format dict."""
        name_dict = self._get_segment('role')
        self._map['role'] = ""

        if len(name_dict['single']) == 0:
            LOG.error('Find 0 role, format error')
            name_dict['single'] = ['unspecified']
            #raise

        for node_role in name_dict['single']:
            self._map['role'] = node_role + ',' + self._map['role']

    def _dict_insert_zone(self):
        """Insert zone into json format dict."""
        name_dict = self._get_segment('zone')
        self._map['zone'] = ""

        if len(name_dict['single']) > 1:
            LOG.error('Format error: more than 2 zone were given.')
            raise

        if len(name_dict['single']) == 0:
            LOG.debug('Use default zone: %s' % FLAGS.default_zone)
            name_dict['single'].append(FLAGS.default_zone)

        for single_zone in name_dict['single']:
            self._map['zone'] = single_zone

    def _dict_insert_auth_key(self):
        """Insert auth key into json format dict."""
        name_dict = self._get_segment('auth_key')
        self._map['auth_key'] = ""

        if len(name_dict['single']) > 1 or \
           len(name_dict['single']) == 0:
            LOG.error('Find 0 or multiple auth_key, format error')
            raise

        for auth_key in name_dict['single']:
            self._map['auth_key'] = auth_key

    def _dict_insert_storage_class(self, storage_type):
        """Try to insert storage class into json format dict."""
        name_dict = self._get_segment(storage_type)
        if len(name_dict['first']) != len(name_dict['second']):
            LOG.error('The first row must be the same as second row')
            raise

        if self._map.get("storage_class", None) is None:
            self._map["storage_class"] = []

        storage_class = {'name': storage_type, 'disk': []}
        for idx in range(0, len(name_dict['first'])):
            one_disk = {"osd": name_dict['first'][idx],
                        "journal": name_dict['second'][idx]}
            storage_class['disk'].append(one_disk)

        self._map["storage_class"].append(storage_class)

    def _dict_insert_sys_info(self):
        """Insert into system info."""
        if not self._map:
            self._map = {}

        self._map["host"] = sys_info.get_hostname()
        self._map["ip"] = sys_info.get_local_ip()
        self._map["id_rsa_pub"] = sys_info.get_rsa_key()

    def _format_server_manifest_to_json(self, check_manifest_tag):
        """Transfer the server manifest file into json format dict."""
        self._map = {}
        if not check_manifest_tag:
            disk_type_list = self._get_disk_type_list()
            for disk_type in disk_type_list:
                self._dict_insert_storage_class(disk_type)

        self._dict_insert_controller_ip()
        self._dict_insert_role()
        self._dict_insert_zone()
        self._dict_insert_auth_key()
        #self._map['drive_num'] = self._number_of_disks()
        self._dict_insert_sys_info()

        # Refer to the same ip
        self._map['cluster_ip'] = self._map['vsm_controller_ip']
        self._map['controller_ip'] = self._map['vsm_controller_ip']
        #self._map['device_type_list'] = disk_type_list
        return self._map

    def _dict_insert_ntp_keys(self):
        self._map["ntp_keys"] = sys_info.get_ntp_keys()

    def _dict_insert_storage_class_c(self):
        """Insert storage_class into json format dict."""
        name_dict = self._get_segment('storage_class')

        if len(name_dict['single']) == 0:
            LOG.error('Format error, at least 1 storage class should be given')
            raise

        self._map['storage_class'] = []
        for storage_class in name_dict['single']:
            self._map['storage_class'].append(storage_class)

    def _dict_insert_periodic_parameters_c(self):
        name_dict = self._get_segment('storage_group_near_full_threshold')
        if len(name_dict['single']) == 0:
            LOG.error('Format error, at least, need set' + \
                      'storage_group_near_full_threshold')
        self._map['storage_group_near_full_threshold'] = name_dict['single']

        name_dict = self._get_segment('storage_group_full_threshold')
        if len(name_dict['single']) == 0:
            LOG.error('Format error, at least, need set' + \
                      'storage_group_full_threshold')
        self._map['storage_group_full_threshold'] = name_dict['single']

    def _dict_insert_zone_c(self):
        """Insert zone into json format dict."""
        name_dict = self._get_segment('zone')
        self._map['zone'] = []

        zone_list = list(name_dict['single'])
        if len(zone_list) != len(name_dict['single']):
            LOG.error("There is exist duplicate zone, please check")

        for zone in zone_list:
            self._map['zone'].append(zone)

        if len(self._map['zone']) == 0:
            self._map['zone'].append(FLAGS.default_zone)

    def _dict_insert_openstack_ip_c(self):
        """ Insert Openstack nova compute and cinder volume nodes'
            ip into json format
        """
        name_dict = self._get_segment('openstack_ip')
        self._map['openstack_ip'] = list()

        nova_set = set(name_dict['single'])
        if len(nova_set) != len(name_dict['single']):
            LOG.error("There is exist duplicate ip, please check")

        for ip in nova_set:
            self._map['openstack_ip'].append(ip)

    def _dict_insert_storage_group_c(self):
        """Insert storage_group into json format dict."""
        name_dict = self._get_segment('storage_group')

        if len(name_dict['first']) == 0:
            LOG.error('Format error, at least 1 storage group should be given')
            raise

        self._map['storage_group'] = []
        for idx in range(0, len(name_dict['first'])):
            one_storage_group = {"name": name_dict['first'][idx],
                                 "friendly_name": name_dict['second'][idx],
                                 "storage_class": name_dict['third'][idx]}
            self._map['storage_group'].append(one_storage_group)

    def _dict_insert_cluster_c(self):
        """Return the segments of cluster name."""
        #return self._get_segment("cluster")
        self._map['cluster'] = self._get_segment("cluster")

    def _dict_insert_ceph_conf(self):
        """ Return the segments of ceph_conf."""
        name_dict = self._get_segment("ceph_conf")
        LOG.info('name_dict===%s'%name_dict)
        self._map['ceph_conf'] = []
        for idx in range(0, len(name_dict['first'])):
            ceph_conf = {'name': name_dict['first'][idx],
                       'default_value': name_dict['second'][idx]}
            self._map['ceph_conf'].append(ceph_conf)
        LOG.info("ceph_conf (key/value) set: %s" % self._map['ceph_conf'])

    def _dict_insert_settings_c(self):
        """ Return the segments of vsm settings."""
        name_dict = self._get_segment("settings")
        #LOG.info("settings (key/value) set: %s" % self._map['settings'])
        self._map['settings'] = []
        for idx in range(0, len(name_dict['first'])):
            setting = {'name': name_dict['first'][idx],
                       'default_value': name_dict['second'][idx]}
            self._map['settings'].append(setting)

    def _dict_insert_cache_tier_defaults_c(self):
        """ Return the segments of cache_tier default value."""
        name_dict = self._get_segment("cache_tier_defaults")
        self._map['cache_tier_defaults'] = []
        for idx in range(0, len(name_dict['first'])):
            setting = {'name': name_dict['first'][idx],
                       'default_value': name_dict['second'][idx]}
            self._map['cache_tier_defaults'].append(setting)

    def _dict_insert_server_list_c(self):
        """Insert server_list into json format dict"""
        name_dict = self._get_segment('server_list')
        if len(name_dict['single']) == 0:
            return

        self._map['server_list'] = []
        for server in name_dict['single']:
            self._map['server_list'].append(server)

    def _dict_insert_ec_profiles_c(self):
        """Return the segments of erasure code profiles"""
        name_dict = self._get_segment("ec_profiles")
        print name_dict
        LOG.info("ec_profiles:%s" % name_dict) 
        print name_dict

        if len(name_dict['first']) == 0:
            # NOTE: we may have to support old version of cluster.manifest file.
            return
            LOG.error('Format error, at least 1 storage group should be given')
            raise

        self._map['ec_profiles'] = []
        for idx in range(0, len(name_dict['first'])):
            one_ec_profile = {"name": name_dict['first'][idx],
                                 "plugin_path": name_dict['second'][idx],
                                 "plugin": name_dict['third'][idx],
                                 "pg_num": name_dict['fourth'][idx],
                                 "plugin_kv_pair": name_dict['fifth'][idx]} 
            self._map['ec_profiles'].append(one_ec_profile)

    def _trans_between_db(self):
        """Write this info for writing into DB."""
        if not self._map:
            return
        info = self._map['cluster']
        self._map['cluster']['name'] = info["cluster_name"]
        self._map['cluster']['management_network'] = info["public_addr"]
        self._map['cluster']['ceph_public_network'] = \
            info["secondary_public_addr"]
        self._map['cluster']['cluster_network'] = info["cluster_addr"]
        self._map['cluster']['file_system'] = info["file_system"]
    def _format_cluster_manifest_to_json(self):
        """Return the json format [dict]."""
        self._map = {}
        self._dict_insert_storage_class_c()
        self._dict_insert_zone_c()
        self._dict_insert_storage_group_c()
        self._dict_insert_cluster_c()
        self._dict_insert_server_list_c()
        self._dict_insert_ntp_keys()
        self._dict_insert_openstack_ip_c()
        self._dict_insert_settings_c()
        self._dict_insert_cache_tier_defaults_c()
        self._dict_insert_ec_profiles_c()
        self._dict_insert_ceph_conf()
        #self._dict_insert_periodic_parameters_c()
        self._trans_between_db()
        return self._map

    def format_to_json(self, check_manifest_tag=False):
        """Transfer the manifest file into json format dict"""
        if self._is_server_manifest:
            return self._format_server_manifest_to_json(check_manifest_tag)
        else:
            return self._format_cluster_manifest_to_json()




# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2014 Intel Inc.
# All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the"License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#  http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

import webob
from webob import exc
import re
import json
from vsm.api import common
from vsm.api.openstack import wsgi
from vsm.api import xmlutil
from vsm import exception
from vsm import flags
from vsm.openstack.common import log as logging
from vsm.api.views import clusters as clusters_views
from vsm.api.views import summary as summary_view
from vsm.openstack.common import jsonutils
from vsm import utils
from vsm import conductor
from vsm import scheduler
from vsm import db
from vsm.agent import cephconfigparser

LOG = logging.getLogger(__name__)

FLAGS = flags.FLAGS

def make_cluster(elem, detailed=False):
    elem.set('id')
    elem.set('name')
    elem.set('file_system')
    elem.set('journal_size')
    elem.set('size')
    elem.set('management_network')
    elem.set('ceph_public_network')
    elem.set('cluster_network')
    elem.set('primary_public_ip_netmask')
    elem.set('secondary_public_ip_netmask')
    elem.set('cluster_ip_netmask')

    if detailed:
        pass

    #xmlutil.make_links(elem, 'links')

cluster_nsmap = {None: xmlutil.XMLNS_V11, 'atom': xmlutil.XMLNS_ATOM}

class ClusterTemplate(xmlutil.TemplateBuilder):
    def construct(self):
        root = xmlutil.TemplateElement('cluster', selector='cluster')
        make_cluster(root, detailed=True)
        return xmlutil.MasterTemplate(root, 1, nsmap=cluster_nsmap)

class ClustersTemplate(xmlutil.TemplateBuilder):
    def construct(self):
        root = xmlutil.TemplateElement('clusters')
        elem = xmlutil.SubTemplateElement(root, 'cluster', selector='clusters')
        make_cluster(elem, detailed=True)
        return xmlutil.MasterTemplate(root, 1, nsmap=cluster_nsmap)

class ClusterController(wsgi.Controller):
    """The Cluster API controller for the OpenStack API."""
    _view_builder_class = clusters_views.ViewBuilder

    def __init__(self, ext_mgr):
        self.conductor_api = conductor.API()
        self.scheduler_api = scheduler.API()
        self.ext_mgr = ext_mgr
        super(ClusterController, self).__init__()

    def _get_server_search_options(self):
        """Return server search options allowed by non-admin."""
        return ('id', 'name', 'public_ip')

    @wsgi.serializers(xml=ClustersTemplate)
    def index(self, req):
        """Get cluster list."""
        search_opts = {}
        search_opts.update(req.GET)
        context = req.environ['vsm.context']
        remove_invalid_options(context,
                                search_opts,
                                self._get_server_search_options)
        clusters = self.conductor_api.get_cluster_list(context)
        #server_list = servers.values()
        #sorted_servers = sorted(server_list,
        #                      key=lambda item: item['id'])

        return self._view_builder.index(clusters)

    #@wsgi.serializers(xml=ClustersTemplate)
    def create(self, req, body):
        """create cluster."""
        LOG.info("CEPH_LOG cluster create body: %s" % body)
        context = req.environ['vsm.context']
        server_list = body['cluster']['servers']
        LOG.info('Begin to call scheduler.createcluster')

        # only and should only one mds
        mds_count = 0
        for server in server_list:
            if server['is_mds'] == True:
                mds_count = mds_count + 1
        if mds_count > 1:
            raise exc.HTTPBadRequest("More than one mds.")

        # only and should only one rgw
        rgw_count = 0
        for server in server_list:
            if server['is_rgw'] == True:
                rgw_count = rgw_count + 1
        if rgw_count > 1:
            raise exc.HTTPBadRequest("More than one rgw.")

        self.scheduler_api.create_cluster(context, server_list)
        return webob.Response(status_int=202)

    def import_ceph_conf(self, req, body=None):
        """
        import_ceph_conf to db and ceph nodes
        """
        LOG.info("CEPH_LOG import_ceph_conf body=%s"%body )
        context = req.environ['vsm.context']
        ceph_conf_path = body["cluster"]["ceph_conf_path"]
        cluster_name = body["cluster"]["cluster_name"]
        cluster = db.cluster_get_by_name(context,cluster_name)
        if cluster:
            ceph_conf_dict_old = cephconfigparser.CephConfigParser(FLAGS.ceph_conf)._parser.as_dict()
            ceph_conf_parser = cephconfigparser.CephConfigParser(ceph_conf_path)._parser
            ceph_conf_dict = ceph_conf_parser.as_dict()
            check_ret = check_ceph_conf(ceph_conf_dict_old,ceph_conf_dict)
            if check_ret:
                return {"message":"%s"%check_ret}
            self.scheduler_api.import_ceph_conf(context,cluster_id=cluster.id,ceph_conf_path=ceph_conf_path)
            return {"message":"Success"}
        else:
            return {"message":"No such cluster which named  %s in DB"%cluster_name}

    def detect_cephconf(self, req, body=None):
        '''

        :param res:
        :param body:
        :return:
        '''
        LOG.info("CEPH_LOG detect_cephconf body=%s"%body )
        context = req.environ['vsm.context']
        ret = self.scheduler_api.detect_cephconf(context,body)
        LOG.info('CEPH_LOG detect_cephconf get ret=%s'%ret)
        return ret

    def detect_crushmap(self, req, body=None):
        '''

        :param res:
        :param body:
        :return:
        '''
        LOG.info("CEPH_LOG detect_crushmap body=%s"%body )
        context = req.environ['vsm.context']
        ret = self.scheduler_api.detect_crushmap(context,body)
        LOG.info('CEPH_LOG detect_crushmap get ret=%s'%ret)
        return ret

    def get_crushmap_tree_data(self, req, body=None):
        '''

        :param res:
        :param body:
        :return:
        '''
        LOG.info("CEPH_LOG get_crushmap_tree_data body=%s"%body )
        context = req.environ['vsm.context']
        ret = self.scheduler_api.get_crushmap_tree_data(context,body)
        LOG.info('CEPH_LOG get_crushmap_tree_data get ret=%s'%ret)
        return ret

    def check_pre_existing_cluster(self,req,body):
        '''

        :param res:
        :param body:
        :return:
        '''
        LOG.info("CEPH_LOG check_pre_existing_cluster body=%s"%body )
        context = req.environ['vsm.context']
        ret = self.scheduler_api.check_pre_existing_cluster(context,body)
        LOG.info('CEPH_LOG check_pre_existing_cluster get ret=%s'%ret)
        return ret

    def import_cluster(self,req,body):
        '''

        :param res:
        :param body:
        :return:
        '''
        LOG.info("CEPH_LOG import_cluster body=%s"%body )
        context = req.environ['vsm.context']
        ret = self.scheduler_api.import_cluster(context,body)
        LOG.info('CEPH_LOG import_cluster get ret=%s'%ret)
        return ret

    def integrate(self, req,body=None):
        """
        integrate an existing ceph cluster
        """
        LOG.info("CEPH_LOG cluster integrate body" )
        context = req.environ['vsm.context']
        #server_list = body['cluster']['servers']
        LOG.info('Begin to call scheduler.integrate_cluster')
        self.scheduler_api.integrate_cluster(context)

    def start_cluster(self, req, body=None):
        """
        start_cluster
        """
        LOG.info("CEPH_LOG start_cluster body=%s"%body )
        context = req.environ['vsm.context']
        cluster_id = body["cluster"]["id"]
        if cluster_id:
            nodes = db.init_node_get_by_cluster_id(context,cluster_id)
        else:
            nodes = db.init_node_get_all(context)
        servers = {"servers":nodes}
        self.scheduler_api.start_cluster(context,servers)
        return {"message":"Success"}

    def stop_cluster(self, req, body=None):
        """
        stop_cluster
        """
        LOG.info("CEPH_LOG stop_cluster body=%s"%body )
        context = req.environ['vsm.context']
        cluster_id = body["cluster"]["id"]
        if cluster_id:
            nodes = db.init_node_get_by_cluster_id(context,cluster_id)
        else:
            nodes = db.init_node_get_all(context)
        servers = {"servers":nodes}
        self.scheduler_api.stop_cluster(context,servers)
        return {"message":"Success"}

    def get_ceph_health_list(self, req):
        context = req.environ['vsm.context']
        ceph_status = self.scheduler_api.get_ceph_health_list(context)
        return {"ceph_status":ceph_status}

    @wsgi.serializers(xml=ClustersTemplate)
    def show(self, req, id):
        """update cluster."""
        LOG.info("CEPH_LOG cluster show id: %s" % id)
        return {"cluster": {"id":1,"name":"2"}}

    @wsgi.serializers(xml=ClustersTemplate)
    def update(self, req, id, body):
        """update cluster."""
        LOG.info("CEPH_LOG cluster update body: %s" % body)
        return {"cluster": {"id":1,"name":"2"}}

    def delete(self, req, id):
        """delete cluster."""
        LOG.info("CEPH_LOG cluster delete id: %s" % id)
        return webob.Response(status_int=202)

    def summary(self, req,cluster_id=None):
        #LOG.info('osd-summary body %s ' % body)
        context = req.environ['vsm.context']
        #TODO: as we have only one cluster for now, the cluster_id
        #has been hardcoded. In furture, client should pass
        # the cluster id by url.
        if cluster_id:
            sum = db.summary_get_by_cluster_id_and_type(context, cluster_id, 'cluster')
        else:
            sum = db.summary_get_by_type_first(context, 'cluster')
        vb = summary_view.ViewBuilder()
        return vb.basic(sum, 'cluster')

    def get_service_list(self, req,cluster_id=None):
        context = req.environ['vsm.context']
        service = db.service_get_all(context)
        return {"services":service}

    def refresh(self, req):
        """
        :param req:
        :return:
        refresh cluster status
        """
        context = req.environ['vsm.context']
        self.scheduler_api.cluster_refresh(context)

def create_resource(ext_mgr):
    return wsgi.Resource(ClusterController(ext_mgr))

def remove_invalid_options(context, search_options, allowed_search_options):
    """Remove search options that are not valid for non-admin API/context."""
    if context.is_admin:
        # Allow all options
        return
    # Otherwise, strip out all unknown options
    unknown_options = [opt for opt in search_options
                       if opt not in allowed_search_options]
    bad_options = ", ".join(unknown_options)
    log_msg = _("Removing options '%(bad_options)s' from query") % locals()
    LOG.debug(log_msg)
    for opt in unknown_options:
        del search_options[opt]

def check_ceph_conf(ceph_conf_dict_old,ceph_conf_dict):
    check_result = ''
    keys_old = ceph_conf_dict_old.keys()
    keys_new = ceph_conf_dict.keys()
    old_osds = [key  for key in keys_old if key.find("osd.")!=-1]
    old_mons = [key  for key in keys_old if key.find("mon.")!=-1]
    old_mds = [key  for key in keys_old if key.find("mds.")!=-1]
    new_osds = [key  for key in keys_new if key.find("osd.")!=-1]
    new_mons = [key  for key in keys_new if key.find("mon.")!=-1]
    new_mds = [key  for key in keys_new if key.find("mds.")!=-1]
    if set(old_osds) != set(new_osds):
        check_result = '%s\n.error:The number of osd is different from the old version'%check_result
    if set(old_mons) != set(new_mons):
        check_result = '%s\n.error:The number of mon is different from the old version'%check_result
    if set(old_mds) != set(new_mds):
        check_result = '%s\n.error:The number of mds is different from the old version'%check_result
    if check_result:
        return check_result
    try:
        osd_field_to_check = ["osd journal","devs","host","cluster addr","public addr"]
        for osd in new_osds:
            for field in osd_field_to_check:
                if ceph_conf_dict[osd][field]!= ceph_conf_dict_old[osd][field]:
                   check_result = '%s\n.error:the value of %s below section %s'%(check_result,field,osd)
        mon_field_to_check = ['mon addr','host']
        for mon in new_mons:
            for field in mon_field_to_check:
                if ceph_conf_dict[mon][field]!= ceph_conf_dict_old[mon][field]:
                   check_result = '%s\n.error:the value of %s below section %s'%(check_result,field,mon)
        mds_field_to_check = ['public addr','host']
        for mds in new_mds:
            for field in mds_field_to_check:
                if ceph_conf_dict[mds][field]!= ceph_conf_dict_old[mds][field]:
                   check_result = '%s\n.error:the value of %s below section %s'%(check_result,field,mds)
    except Exception,e:
        check_result = '%s\n.error:KeyError :%s'%(check_result,e)
    return check_result



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2014 Intel
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Agent Service
"""

import os
import json
import re
import time
from vsm import exception
from vsm import flags
from vsm.openstack.common import log as logging
from vsm import utils

LOG = logging.getLogger(__name__)
FLAGS = flags.FLAGS

def set_keyring_admin(keyring_admin):
    """Set keyring admin file."""
    if not os.path.exists(FLAGS.keyring_admin):
        #open(FLAGS.keyring_admin, 'w').write(keyring_admin)
        utils.write_file_as_root(FLAGS.keyring_admin, keyring_admin, "w")
    else:
        LOG.info('Error keyring file exists')
        LOG.info('Back up & cover the old version')
        old_content = open(FLAGS.keyring_admin, 'r').read()
        # if not the same. cmp('a', 'a') == 0
        if cmp(keyring_admin.strip(), old_content.strip()):
            bfx = time.asctime().replace(' ','_').replace(':','_')
            old_fname = FLAGS.keyring_admin + bfx
            #open(old_fname, 'w').write(old_content)
            utils.write_file_as_root(old_fname, old_content, "w")
            # write the newer version.
            #open(FLAGS.keyring_admin, 'w').write(keyring_admin)
            utils.write_file_as_root(FLAGS.keyring_admin, keyring_admin, "w")
        else:
            LOG.info('Have the same content, pass')

def is_vsm_ok():
    """Check vsm.conf is configured."""
    vcf = file(FLAGS.vsm_config).read()
    apf = file(FLAGS.api_paste_config).read()

    if re.search("MYSQL_VSM_PASSWORD", vcf)\
       or re.search("MYSQL_VSM_USER", vcf)\
       or re.search("RABBITMQ_HOST", vcf)\
       or re.search("RABBITMQ_PASSWORD", vcf)\
       or re.search("RABBITMQ_PORT", vcf)\
       or re.search("SERVICE_TENANT_NAME", apf)\
       or re.search("SERVICE_USER", apf)\
       or re.search("SERVICE_PASSWORD", apf):
        #utils.execute("rm", "-rf", "/root/.ssh/", run_as_root=True)
        #os.popen('rm -rf /root/.ssh/')
        return False
    return True

def set_vsm_conf(recv):
    """Write vsm conf files."""
    if recv is None:
        return False

    old_recv = recv
    try:
        recv = json.loads(recv)
    except TypeError:
        recv = old_recv
        LOG.info('Maybe recv is not json.')

    if not recv.get("api-paste.ini", None):
        LOG.error('Can not find content of api-paste.ini')
        raise exception.NotFound()

    if not recv.get("vsm.conf", None):
        LOG.error('Can not find content of vsm.conf')
        raise exception.NotFound()

    files = ["api-paste.ini", "vsm.conf"]
    for conf in files:
        conf_path = FLAGS.vsm_config_path + conf
        #conf_file = open(conf_path, 'w')
        #conf_file.write(recv[conf])
        utils.write_file_as_root(conf_path, recv[conf], "w")
        #conf_file.close()

    return True



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2010 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utility methods for working with WSGI servers."""

import errno
import os
import socket
import ssl
import sys
import time

import eventlet
import eventlet.wsgi
import greenlet
from oslo.config import cfg
from paste import deploy
import routes.middleware
import webob.dec
import webob.exc

from vsm import exception
from vsm import flags
from vsm.openstack.common import log as logging
from vsm import utils

socket_opts = [
    cfg.IntOpt('backlog',
               default=4096,
               help="Number of backlog requests to configure the socket with"),
    cfg.IntOpt('tcp_keepidle',
               default=600,
               help="Sets the value of TCP_KEEPIDLE in seconds for each "
                    "server socket. Not supported on OS X."),
    cfg.StrOpt('ssl_ca_file',
               default=None,
               help="CA certificate file to use to verify "
                    "connecting clients"),
    cfg.StrOpt('ssl_cert_file',
               default=None,
               help="Certificate file to use when starting "
                    "the server securely"),
    cfg.StrOpt('ssl_key_file',
               default=None,
               help="Private key file to use when starting "
                    "the server securely"),
]

CONF = cfg.CONF
CONF.register_opts(socket_opts)

FLAGS = flags.FLAGS
LOG = logging.getLogger(__name__)

class Server(object):
    """Server class to manage a WSGI server, serving a WSGI application."""

    default_pool_size = 1000

    def __init__(self, name, app, host=None, port=None, pool_size=None,
                 protocol=eventlet.wsgi.HttpProtocol):
        """Initialize, but do not start, a WSGI server.

        :param name: Pretty name for logging.
        :param app: The WSGI application to serve.
        :param host: IP address to serve the application.
        :param port: Port number to server the application.
        :param pool_size: Maximum number of eventlets to spawn concurrently.
        :returns: None

        """
        self.name = name
        self.app = app
        self._host = host or "0.0.0.0"
        self._port = port or 0
        self._server = None
        self._socket = None
        self._protocol = protocol
        self._pool = eventlet.GreenPool(pool_size or self.default_pool_size)
        self._logger = logging.getLogger("eventlet.wsgi.server")
        self._wsgi_logger = logging.WritableLogger(self._logger)

    def _get_socket(self, host, port, backlog):
        bind_addr = (host, port)
        # TODO(dims): eventlet's green dns/socket module does not actually
        # support IPv6 in getaddrinfo(). We need to get around this in the
        # future or monitor upstream for a fix
        try:
            info = socket.getaddrinfo(bind_addr[0],
                                      bind_addr[1],
                                      socket.AF_UNSPEC,
                                      socket.SOCK_STREAM)[0]
            family = info[0]
            bind_addr = info[-1]
        except Exception:
            family = socket.AF_INET

        cert_file = CONF.ssl_cert_file
        key_file = CONF.ssl_key_file
        ca_file = CONF.ssl_ca_file
        use_ssl = cert_file or key_file

        if cert_file and not os.path.exists(cert_file):
            raise RuntimeError(_("Unable to find cert_file : %s") % cert_file)

        if ca_file and not os.path.exists(ca_file):
            raise RuntimeError(_("Unable to find ca_file : %s") % ca_file)

        if key_file and not os.path.exists(key_file):
            raise RuntimeError(_("Unable to find key_file : %s") % key_file)

        if use_ssl and (not cert_file or not key_file):
            raise RuntimeError(_("When running server in SSL mode, you must "
                                 "specify both a cert_file and key_file "
                                 "option value in your configuration file"))

        def wrap_ssl(sock):
            ssl_kwargs = {
                'server_side': True,
                'certfile': cert_file,
                'keyfile': key_file,
                'cert_reqs': ssl.CERT_NONE,
            }

            if CONF.ssl_ca_file:
                ssl_kwargs['ca_certs'] = ca_file
                ssl_kwargs['cert_reqs'] = ssl.CERT_REQUIRED

            return ssl.wrap_socket(sock, **ssl_kwargs)

        sock = None
        retry_until = time.time() + 30
        while not sock and time.time() < retry_until:
            try:
                sock = eventlet.listen(bind_addr,
                                       backlog=backlog,
                                       family=family)
                if use_ssl:
                    sock = wrap_ssl(sock)

            except socket.error, err:
                if err.args[0] != errno.EADDRINUSE:
                    raise
                eventlet.sleep(0.1)
        if not sock:
            raise RuntimeError(_("Could not bind to %(host)s:%(port)s "
                               "after trying for 30 seconds") %
                               {'host': host, 'port': port})
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # sockets can hang around forever without keepalive
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

        # This option isn't available in the OS X version of eventlet
        if hasattr(socket, 'TCP_KEEPIDLE'):
            sock.setsockopt(socket.IPPROTO_TCP,
                            socket.TCP_KEEPIDLE,
                            CONF.tcp_keepidle)

        return sock

    def _start(self):
        """Run the blocking eventlet WSGI server.

        :returns: None

        """
        eventlet.wsgi.server(self._socket,
                             self.app,
                             protocol=self._protocol,
                             custom_pool=self._pool,
                             log=self._wsgi_logger)

    def start(self, backlog=128):
        """Start serving a WSGI application.

        :param backlog: Maximum number of queued connections.
        :returns: None
        :raises: vsm.exception.InvalidInput

        """
        if backlog < 1:
            raise exception.InvalidInput(
                reason='The backlog must be more than 1')

        self._socket = self._get_socket(self._host,
                                        self._port,
                                        backlog=backlog)
        self._server = eventlet.spawn(self._start)
        (self._host, self._port) = self._socket.getsockname()[0:2]
        LOG.info(_("Started %(name)s on %(_host)s:%(_port)s") % self.__dict__)

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    def stop(self):
        """Stop this server.

        This is not a very nice action, as currently the method by which a
        server is stopped is by killing its eventlet.

        :returns: None

        """
        LOG.info(_("Stopping WSGI server."))
        self._server.kill()

    def wait(self):
        """Block, until the server has stopped.

        Waits on the server's eventlet to finish, then returns.

        :returns: None

        """
        try:
            self._server.wait()
        except greenlet.GreenletExit:
            LOG.info(_("WSGI server has stopped."))

class Request(webob.Request):
    pass

class Application(object):
    """Base WSGI application wrapper. Subclasses need to implement __call__."""

    @classmethod
    def factory(cls, global_config, **local_config):
        """Used for paste app factories in paste.deploy config files.

        Any local configuration (that is, values under the [app:APPNAME]
        section of the paste config) will be passed into the `__init__` method
        as kwargs.

        A hypothetical configuration would look like:

            [app:wadl]
            latest_version = 1.3
            paste.app_factory = vsm.api.fancy_api:Wadl.factory

        which would result in a call to the `Wadl` class as

            import vsm.api.fancy_api
            fancy_api.Wadl(latest_version='1.3')

        You could of course re-implement the `factory` method in subclasses,
        but using the kwarg passing it shouldn't be necessary.

        """
        return cls(**local_config)

    def __call__(self, environ, start_response):
        r"""Subclasses will probably want to implement __call__ like this:

        @webob.dec.wsgify(RequestClass=Request)
        def __call__(self, req):
          # Any of the following objects work as responses:

          # Option 1: simple string
          res = 'message\n'

          # Option 2: a nicely formatted HTTP exception page
          res = exc.HTTPForbidden(detail='Nice try')

          # Option 3: a webob Response object (in case you need to play with
          # headers, or you want to be treated like an iterable, or or or)
          res = Response();
          res.app_iter = open('somefile')

          # Option 4: any wsgi app to be run next
          res = self.application

          # Option 5: you can get a Response object for a wsgi app, too, to
          # play with headers etc
          res = req.get_response(self.application)

          # You can then just return your response...
          return res
          # ... or set req.response and return None.
          req.response = res

        See the end of http://pythonpaste.org/webob/modules/dec.html
        for more info.

        """
        raise NotImplementedError(_('You must implement __call__'))

class Middleware(Application):
    """Base WSGI middleware.

    These classes require an application to be
    initialized that will be called next.  By default the middleware will
    simply call its wrapped app, or you can override __call__ to customize its
    behavior.

    """

    @classmethod
    def factory(cls, global_config, **local_config):
        """Used for paste app factories in paste.deploy config files.

        Any local configuration (that is, values under the [filter:APPNAME]
        section of the paste config) will be passed into the `__init__` method
        as kwargs.

        A hypothetical configuration would look like:

            [filter:analytics]
            redis_host = 127.0.0.1
            paste.filter_factory = vsm.api.analytics:Analytics.factory

        which would result in a call to the `Analytics` class as

            import vsm.api.analytics
            analytics.Analytics(app_from_paste, redis_host='127.0.0.1')

        You could of course re-implement the `factory` method in subclasses,
        but using the kwarg passing it shouldn't be necessary.

        """
        def _factory(app):
            return cls(app, **local_config)
        return _factory

    def __init__(self, application):
        self.application = application

    def process_request(self, req):
        """Called on each request.

        If this returns None, the next application down the stack will be
        executed. If it returns a response then that response will be returned
        and execution will stop here.

        """
        return None

    def process_response(self, response):
        """Do whatever you'd like to the response."""
        return response

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, req):
        response = self.process_request(req)
        if response:
            return response
        response = req.get_response(self.application)
        return self.process_response(response)

class Debug(Middleware):
    """Helper class for debugging a WSGI application.

    Can be inserted into any WSGI application chain to get information
    about the request and response.

    """

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, req):
        print ('*' * 40) + ' REQUEST ENVIRON'
        for key, value in req.environ.items():
            print key, '=', value
        print
        resp = req.get_response(self.application)

        print ('*' * 40) + ' RESPONSE HEADERS'
        for (key, value) in resp.headers.iteritems():
            print key, '=', value
        print

        resp.app_iter = self.print_generator(resp.app_iter)

        return resp

    @staticmethod
    def print_generator(app_iter):
        """Iterator that prints the contents of a wrapper string."""
        print ('*' * 40) + ' BODY'
        for part in app_iter:
            sys.stdout.write(part)
            sys.stdout.flush()
            yield part
        print

class Router(object):
    """WSGI middleware that maps incoming requests to WSGI apps."""

    def __init__(self, mapper):
        """Create a router for the given routes.Mapper.

        Each route in `mapper` must specify a 'controller', which is a
        WSGI app to call.  You'll probably want to specify an 'action' as
        well and have your controller be an object that can route
        the request to the action-specific method.

        Examples:
          mapper = routes.Mapper()
          sc = ServerController()

          # Explicit mapping of one route to a controller+action
          mapper.connect(None, '/svrlist', controller=sc, action='list')

          # Actions are all implicitly defined
          mapper.resource('server', 'servers', controller=sc)

          # Pointing to an arbitrary WSGI app.  You can specify the
          # {path_info:.*} parameter so the target app can be handed just that
          # section of the URL.
          mapper.connect(None, '/v1.0/{path_info:.*}', controller=BlogApp())

        """
        self.map = mapper
        self._router = routes.middleware.RoutesMiddleware(self._dispatch,
                                                          self.map)

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, req):
        """Route the incoming request to a controller based on self.map.

        If no match, return a 404.

        """
        return self._router

    @staticmethod
    @webob.dec.wsgify(RequestClass=Request)
    def _dispatch(req):
        """Dispatch the request to the appropriate controller.

        Called by self._router after matching the incoming request to a route
        and putting the information into req.environ.  Either returns 404
        or the routed WSGI app's response.

        """
        match = req.environ['wsgiorg.routing_args'][1]
        if not match:
            return webob.exc.HTTPNotFound()
        app = match['controller']
        return app

class Loader(object):
    """Used to load WSGI applications from paste configurations."""

    def __init__(self, config_path=None):
        """Initialize the loader, and attempt to find the config.

        :param config_path: Full or relative path to the paste config.
        :returns: None

        """
        config_path = config_path or FLAGS.api_paste_config
        self.config_path = utils.find_config(config_path)

    def load_app(self, name):
        """Return the paste URLMap wrapped WSGI application.

        :param name: Name of the application to load.
        :returns: Paste URLMap object wrapping the requested application.
        :raises: `vsm.exception.PasteAppNotFound`

        """
        try:
            return deploy.loadapp("config:%s" % self.config_path, name=name)
        except LookupError as err:
            LOG.error(err)
            raise exception.PasteAppNotFound(name=name, path=self.config_path)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2014 Intel Inc.
# All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the"License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#  http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

import webob
from webob import exc
from vsm.api.openstack import wsgi
from vsm.api import xmlutil
from vsm import flags
from vsm.openstack.common import log as logging
from vsm.api.views import monitors as mon_views
from vsm.api.views import summary as sum_views
from vsm import conductor
from vsm import scheduler
from vsm import db
from vsm import exception

LOG = logging.getLogger(__name__)

FLAGS = flags.FLAGS

def make_monitor(elem, detailed=False):
    elem.set('id')
    elem.set('name')
    elem.set('address')
    elem.set('health')
    elem.set('detail')
    elem.set('skew')
    elem.set('latency')
    elem.set('kb_total')
    elem.set('kb_used')
    elem.set('kb_avail')
    elem.set('percent_avail')

    if detailed:
        pass

mon = make_monitor

mon_nsmap = {None: xmlutil.XMLNS_V11, 'atom': xmlutil.XMLNS_ATOM}

class MonitorTemplate(xmlutil.TemplateBuilder):
    def construct(self):
        root = xmlutil.TemplateElement('mon', selector='mon')
        make_monitor(root, detailed=True)
        return xmlutil.MasterTemplate(root, 1, nsmap=mon_nsmap)

class MonitorsTemplate(xmlutil.TemplateBuilder):
    def construct(self):
        root = xmlutil.TemplateElement('mons')
        elem = xmlutil.SubTemplateElement(root, 'mon', selector='mons')
        make_monitor(elem, detailed=True)
        return xmlutil.MasterTemplate(root, 1, nsmap=mon_nsmap)

class Controller(wsgi.Controller):
    """The Monitor API controller for the OpenStack API."""
    _view_builder_class = mon_views.ViewBuilder

    def __init__(self, ext_mgr):
        super(Controller, self).__init__()
        self.conductor_api = conductor.API()
        self.scheduler_api = scheduler.API()

    @wsgi.serializers(xml=MonitorsTemplate)
    def show(self, req, id):
        """Return data about the given mon."""
        context = req.environ['vsm.context']

        try:
            mon = db.monitor_get(context, id)
            error = self.conductor_api.ceph_error(context)
            if error:
                mon['health'] = error

        except exception.NotFound:
            raise exc.HTTPNotFound()

        return self._view_builder.show(mon)

    @wsgi.serializers(xml=MonitorsTemplate)
    def index(self, req):
        """Get mon list."""
        #search_opts = {}
        #search_opts.update(req.GET)
        context = req.environ['vsm.context']

        mons = db.monitor_get_all(context)
        error = self.conductor_api.ceph_error(context)
        if error:
            for mon in mons:
                mons['health'] = error

        LOG.info('vsm/api/v1/mon.py mons:%s' % mons)

        return self._view_builder.index(mons)

    @wsgi.serializers(xml=MonitorsTemplate)
    def detail(self, req):

        """Get mon list."""
        #search_opts = {}
        #search_opts.update(req.GET)
        context = req.environ['vsm.context']
        #remove_invalid_options(context,
        #                        search_opts,
        #                        self._get_zone_search_options)
        #zones = self.conductor_api.get_zone_list(context)
        mons = db.monitor_get_all(context)
        LOG.info('vsm/api/v1/mons.py detailed mons:%s' % mons)

        return self._view_builder.detail(mons)

    @wsgi.response(202)
    def delete(self, req, id):
        """delete a mon in db."""
        context = req.environ['vsm.context']
        mon = db.monitor_get(context, id)

        db.monitor_destroy(context, mon.get('name'))

    @wsgi.response(202)
    @wsgi.action('restart')
    def _action_restart(self, req, id, body):
        context = req.environ['vsm.context']
        LOG.info("action_restart")
        self.scheduler_api.monitor_restart(context, id)
        return webob.Response(status_int=202)

    def summary(self, req, cluster_id=None):
        LOG.info('mon-summary.')
        context = req.environ['vsm.context']
        if cluster_id:
            sum = db.summary_get_by_cluster_id_and_type(context, cluster_id, 'mon')
        else:
            sum = db.summary_get_by_type_first(context, 'mon')
        #sum = db.summary_get_by_cluster_id_and_type(context, 1, 'mon')
        return sum_views.ViewBuilder().basic(sum, 'monitor')

def create_resource(ext_mgr):
    return wsgi.Resource(Controller(ext_mgr))

#def remove_invalid_options(context, search_options, allowed_search_options):
#    """Remove search options that are not valid for non-admin API/context."""
#    if context.is_admin:
#        # Allow all options
#        return
#    # Otherwise, strip out all unknown options
#    unknown_options = [opt for opt in search_options
#                       if opt not in allowed_search_options]
#    bad_options = ", ".join(unknown_options)
#    log_msg = _("Removing options '%(bad_options)s' from query") % locals()
#    LOG.debug(log_msg)
#    for opt in unknown_options:
#        del search_options[opt]



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2011 OpenStack LLC.
# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import webob.dec
import webob.exc

import vsm.api.openstack
from vsm.api.openstack import wsgi
from vsm.api import xmlutil
from vsm import exception
from vsm import flags
from vsm.openstack.common import exception as common_exception
from vsm.openstack.common import importutils
from vsm.openstack.common import log as logging
import vsm.policy

LOG = logging.getLogger(__name__)
FLAGS = flags.FLAGS

class ExtensionDescriptor(object):
    """Base class that defines the contract for extensions.

    Note that you don't have to derive from this class to have a valid
    extension; it is purely a convenience.

    """

    # The name of the extension, e.g., 'Fox In Socks'
    name = None

    # The alias for the extension, e.g., 'FOXNSOX'
    alias = None

    # Description comes from the docstring for the class

    # The XML namespace for the extension, e.g.,
    # 'http://www.fox.in.socks/api/ext/pie/v1.0'
    namespace = None

    # The timestamp when the extension was last updated, e.g.,
    # '2011-01-22T13:25:27-06:00'
    updated = None

    def __init__(self, ext_mgr):
        """Register extension with the extension manager."""

        ext_mgr.register(self)

    def get_resources(self):
        """List of extensions.ResourceExtension extension objects.

        Resources define new nouns, and are accessible through URLs.

        """
        resources = []
        return resources

    def get_controller_extensions(self):
        """List of extensions.ControllerExtension extension objects.

        Controller extensions are used to extend existing controllers.
        """
        controller_exts = []
        return controller_exts

    @classmethod
    def nsmap(cls):
        """Synthesize a namespace map from extension."""

        # Start with a base nsmap
        nsmap = ext_nsmap.copy()

        # Add the namespace for the extension
        nsmap[cls.alias] = cls.namespace

        return nsmap

    @classmethod
    def xmlname(cls, name):
        """Synthesize element and attribute names."""

        return '{%s}%s' % (cls.namespace, name)

def make_ext(elem):
    elem.set('name')
    elem.set('namespace')
    elem.set('alias')
    elem.set('updated')

    desc = xmlutil.SubTemplateElement(elem, 'description')
    desc.text = 'description'

    xmlutil.make_links(elem, 'links')

ext_nsmap = {None: xmlutil.XMLNS_COMMON_V10, 'atom': xmlutil.XMLNS_ATOM}

class ExtensionTemplate(xmlutil.TemplateBuilder):
    def construct(self):
        root = xmlutil.TemplateElement('extension', selector='extension')
        make_ext(root)
        return xmlutil.MasterTemplate(root, 1, nsmap=ext_nsmap)

class ExtensionsTemplate(xmlutil.TemplateBuilder):
    def construct(self):
        root = xmlutil.TemplateElement('extensions')
        elem = xmlutil.SubTemplateElement(root, 'extension',
                                          selector='extensions')
        make_ext(elem)
        return xmlutil.MasterTemplate(root, 1, nsmap=ext_nsmap)

class ExtensionsResource(wsgi.Resource):

    def __init__(self, extension_manager):
        self.extension_manager = extension_manager
        super(ExtensionsResource, self).__init__(None)

    def _translate(self, ext):
        ext_data = {}
        ext_data['name'] = ext.name
        ext_data['alias'] = ext.alias
        ext_data['description'] = ext.__doc__
        ext_data['namespace'] = ext.namespace
        ext_data['updated'] = ext.updated
        ext_data['links'] = []  # TODO(dprince): implement extension links
        return ext_data

    @wsgi.serializers(xml=ExtensionsTemplate)
    def index(self, req):
        extensions = []
        for _alias, ext in self.extension_manager.extensions.iteritems():
            extensions.append(self._translate(ext))
        return dict(extensions=extensions)

    @wsgi.serializers(xml=ExtensionTemplate)
    def show(self, req, id):
        try:
            # NOTE(dprince): the extensions alias is used as the 'id' for show
            ext = self.extension_manager.extensions[id]
        except KeyError:
            raise webob.exc.HTTPNotFound()

        return dict(extension=self._translate(ext))

    def delete(self, req, id):
        raise webob.exc.HTTPNotFound()

    def create(self, req):
        raise webob.exc.HTTPNotFound()

class ExtensionManager(object):
    """Load extensions from the configured extension path.

    See vsm/tests/api/extensions/foxinsocks/extension.py for an
    example extension implementation.

    """

    def __init__(self):
        LOG.audit(_('Initializing extension manager.'))

        self.cls_list = FLAGS.vsmapi_storage_extension
        self.extensions = {}
        self._load_extensions()

    def is_loaded(self, alias):
        return alias in self.extensions

    def register(self, ext):
        # Do nothing if the extension doesn't check out
        if not self._check_extension(ext):
            return

        alias = ext.alias
        LOG.audit(_('Loaded extension: %s'), alias)

        if alias in self.extensions:
            raise exception.Error("Found duplicate extension: %s" % alias)
        self.extensions[alias] = ext

    def get_resources(self):
        """Returns a list of ResourceExtension objects."""

        resources = []
        resources.append(ResourceExtension('extensions',
                                           ExtensionsResource(self)))

        for ext in self.extensions.values():
            try:
                resources.extend(ext.get_resources())
            except AttributeError:
                # NOTE(dprince): Extension aren't required to have resource
                # extensions
                pass
        return resources

    def get_controller_extensions(self):
        """Returns a list of ControllerExtension objects."""
        controller_exts = []
        for ext in self.extensions.values():
            try:
                get_ext_method = ext.get_controller_extensions
            except AttributeError:
                # NOTE(Vek): Extensions aren't required to have
                # controller extensions
                continue
            controller_exts.extend(get_ext_method())
        return controller_exts

    def _check_extension(self, extension):
        """Checks for required methods in extension objects."""
        try:
            LOG.debug(_('Ext name: %s'), extension.name)
            LOG.debug(_('Ext alias: %s'), extension.alias)
            LOG.debug(_('Ext description: %s'),
                      ' '.join(extension.__doc__.strip().split()))
            LOG.debug(_('Ext namespace: %s'), extension.namespace)
            LOG.debug(_('Ext updated: %s'), extension.updated)
        except AttributeError as ex:
            LOG.exception(_("Exception loading extension: %s"), unicode(ex))
            return False

        return True

    def load_extension(self, ext_factory):
        """Execute an extension factory.

        Loads an extension.  The 'ext_factory' is the name of a
        callable that will be imported and called with one
        argument--the extension manager.  The factory callable is
        expected to call the register() method at least once.
        """

        LOG.debug(_("Loading extension %s"), ext_factory)

        # Load the factory
        factory = importutils.import_class(ext_factory)

        # Call it
        LOG.debug(_("Calling extension factory %s"), ext_factory)
        factory(self)

    def _load_extensions(self):
        """Load extensions specified on the command line."""

        extensions = list(self.cls_list)

        # NOTE(thingee): Backwards compat for the old extension loader path.
        # We can drop this post-grizzly in the H release.
        old_contrib_path = ('vsm.api.openstack.storage.contrib.'
                            'standard_extensions')
        new_contrib_path = 'vsm.api.contrib.standard_extensions'
        if old_contrib_path in extensions:
            LOG.warn(_('vsmapi_storage_extension is set to deprecated path: %s'),
                     old_contrib_path)
            LOG.warn(_('Please set your flag or vsm.conf settings for '
                       'vsmapi_storage_extension to: %s'), new_contrib_path)
            extensions = [e.replace(old_contrib_path, new_contrib_path)
                          for e in extensions]

        for ext_factory in extensions:
            try:
                self.load_extension(ext_factory)
            except Exception as exc:
                LOG.warn(_('Failed to load extension %(ext_factory)s: '
                           '%(exc)s') % locals())

class ControllerExtension(object):
    """Extend core controllers of vsm OpenStack API.

    Provide a way to extend existing vsm OpenStack API core
    controllers.
    """

    def __init__(self, extension, collection, controller):
        self.extension = extension
        self.collection = collection
        self.controller = controller

class ResourceExtension(object):
    """Add top level resources to the OpenStack API in vsm."""

    def __init__(self, collection, controller, parent=None,
                 collection_actions=None, member_actions=None,
                 custom_routes_fn=None):
        if not collection_actions:
            collection_actions = {}
        if not member_actions:
            member_actions = {}
        self.collection = collection
        self.controller = controller
        self.parent = parent
        self.collection_actions = collection_actions
        self.member_actions = member_actions
        self.custom_routes_fn = custom_routes_fn

def load_standard_extensions(ext_mgr, logger, path, package, ext_list=None):
    """Registers all standard API extensions."""

    # Walk through all the modules in our directory...
    our_dir = path[0]
    for dirpath, dirnames, filenames in os.walk(our_dir):
        # Compute the relative package name from the dirpath
        relpath = os.path.relpath(dirpath, our_dir)
        if relpath == '.':
            relpkg = ''
        else:
            relpkg = '.%s' % '.'.join(relpath.split(os.sep))

        # Now, consider each file in turn, only considering .py files
        for fname in filenames:
            root, ext = os.path.splitext(fname)

            # Skip __init__ and anything that's not .py
            if ext != '.py' or root == '__init__':
                continue

            # Try loading it
            classname = "%s%s" % (root[0].upper(), root[1:])
            classpath = ("%s%s.%s.%s" %
                         (package, relpkg, root, classname))

            if ext_list is not None and classname not in ext_list:
                logger.debug("Skipping extension: %s" % classpath)
                continue

            try:
                ext_mgr.load_extension(classpath)
            except Exception as exc:
                logger.warn(_('Failed to load extension %(classpath)s: '
                              '%(exc)s') % locals())

        # Now, let's consider any subdirectories we may have...
        subdirs = []
        for dname in dirnames:
            # Skip it if it does not have __init__.py
            if not os.path.exists(os.path.join(dirpath, dname,
                                               '__init__.py')):
                continue

            # If it has extension(), delegate...
            ext_name = ("%s%s.%s.extension" %
                        (package, relpkg, dname))
            try:
                ext = importutils.import_class(ext_name)
            except common_exception.NotFound:
                # extension() doesn't exist on it, so we'll explore
                # the directory for ourselves
                subdirs.append(dname)
            else:
                try:
                    ext(ext_mgr)
                except Exception as exc:
                    logger.warn(_('Failed to load extension %(ext_name)s: '
                                  '%(exc)s') % locals())

        # Update the list of directories we'll explore...
        dirnames[:] = subdirs

def extension_authorizer(api_name, extension_name):
    def authorize(context, target=None):
        if target is None:
            target = {'project_id': context.project_id,
                      'user_id': context.user_id}
        action = '%s_extension:%s' % (api_name, extension_name)
        vsm.policy.enforce(context, action, target)
    return authorize

def soft_extension_authorizer(api_name, extension_name):
    hard_authorize = extension_authorizer(api_name, extension_name)

    def authorize(context):
        try:
            hard_authorize(context)
            return True
        except exception.NotAuthorized:
            return False
    return authorize



# coding=utf-8


import diamond.collector
import commands
try:
    import json
except ImportError:
    import simplejson as json
import glob
import os
import subprocess

def flatten_dictionary(input, sep='.', prefix=None):
    """Produces iterator of pairs where the first value is
    the joined key names and the second value is the value
    associated with the lowest level key. For example::
    {'a': {'b': 10},
    'c': 20,
    }
    produces::
    [('a.b', 10), ('c', 20)]
    """
    for name, value in sorted(input.items()):
        fullname = sep.join(filter(None, [prefix, name]))
    if isinstance(value, dict):
        for result in flatten_dictionary(value, sep, fullname):
            yield result
    else:
        yield (fullname, value)

class CephMetricsCollector(diamond.collector.Collector):

    def get_default_config_help(self):
        config_help = super(CephMetricsCollector, self).get_default_config_help()
        config_help.update({
        'socket_path': 'The location of the ceph monitoring sockets.'
        ' Defaults to "/var/run/ceph"',
        'socket_prefix': 'The first part of all socket names.'
        ' Defaults to "ceph-"',
        'socket_ext': 'Extension for socket filenames.'
        ' Defaults to "asok"',
        'ceph_binary': 'Path to "ceph" executable. '
        'Defaults to /usr/bin/ceph.',
        })
        return config_help

    def get_default_config(self):
        """
        Returns the default collector settings
        """
        config = super(CephMetricsCollector, self).get_default_config()
        config.update({
        'socket_path': '/var/run/ceph',
        'socket_prefix': 'ceph-',
        'socket_ext': 'asok',
        'ceph_binary': '/usr/bin/ceph',
        })
        return config

    def _get_socket_paths(self,type='all'):
        """Return a sequence of paths to sockets for communicating
        with ceph daemons.
        """
        if type == 'all':
            socket_pattern = os.path.join(self.config['socket_path'],
            (self.config['socket_prefix']
            + '*.' + self.config['socket_ext']))
        else:
            socket_pattern = os.path.join(self.config['socket_path'],
            (self.config['socket_prefix']
            + '%s.*.'%type + self.config['socket_ext']))
        return glob.glob(socket_pattern)

    def _get_counter_prefix_from_socket_name(self, name):
        """Given the name of a UDS socket, return the prefix
        for counters coming from that source.
        """
        base = os.path.splitext(os.path.basename(name))[0]
        if base.startswith(self.config['socket_prefix']):
            base = base[len(self.config['socket_prefix']):]
        return 'ceph.' + base

    def _get_stats_from_socket(self, name):
        """Return the parsed JSON data returned when ceph is told to
        dump the stats from the named socket.
        In the event of an error error, the exception is logged, and
        an empty result set is returned.
        """
        try:
            json_blob = subprocess.check_output(
            [self.config['ceph_binary'],
            '--admin-daemon',
            name,
            'perf',
            'dump',
            ])
        except subprocess.CalledProcessError, err:
            self.log.info('Could not get stats from %s: %s',name, err)
            self.log.exception('Could not get stats from %s' % name)
            return {}
        try:
            json_data = json.loads(json_blob)
        except Exception, err:
            self.log.info('Could not parse stats from %s: %s',name, err)
            self.log.exception('Could not parse stats from %s' % name)
            return {}
        return json_data



    def _publish_stats(self, counter_prefix, stats):
        """Given a stats dictionary from _get_stats_from_socket,
        publish the individual values.
        """
        for stat_name, stat_value in flatten_dictionary(
        stats,
        prefix=counter_prefix,
        ):
            self.publish_gauge(stat_name, stat_value)

    def collect(self):
        """
        Collect stats
        """
        for path in self._get_socket_paths(type='osd'):
            self.log.debug('checking %s', path)
            instance_name = path.split('.')[1]
            osd_perf_value_dict = self._get_stats_from_socket(path)['osd']
            metrics = {
                    "osd%s.ops_r"%instance_name: osd_perf_value_dict['op_r'],
                    "osd%s.ops_w"%instance_name: osd_perf_value_dict['op_w'],
                    "osd%s.ops_rw"%instance_name: osd_perf_value_dict['op_rw'],
                    "osd%s.latency_r"%instance_name: osd_perf_value_dict['op_r_latency']['avgcount'] and osd_perf_value_dict['op_r_latency']['sum']/osd_perf_value_dict['op_r_latency']['avgcount'] or 0,
                    "osd%s.latency_w"%instance_name: osd_perf_value_dict['op_w_latency']['avgcount'] and osd_perf_value_dict['op_w_latency']['sum']/osd_perf_value_dict['op_w_latency']['avgcount'] or 0,
                    "osd%s.latency_rw"%instance_name: osd_perf_value_dict['op_rw_latency']['avgcount'] and osd_perf_value_dict['op_rw_latency']['sum']/osd_perf_value_dict['op_rw_latency']['avgcount'] or 0,
                    "osd%s.bandwidth_in"%instance_name: osd_perf_value_dict['op_in_bytes'],
                    "osd%s.bandwidth_out"%instance_name: osd_perf_value_dict['op_out_bytes'],
                }
            for key,value in metrics.items():
                self.publish(key, value)
        return




# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ConfigParser
import logging
import logging.handlers
import os
import string

from vsm.openstack.common.rootwrap import filters

class NoFilterMatched(Exception):
    """This exception is raised when no filter matched."""
    pass

class FilterMatchNotExecutable(Exception):
    """
    This exception is raised when a filter matched but no executable was
    found.
    """
    def __init__(self, match=None, **kwargs):
        self.match = match

class RootwrapConfig(object):

    def __init__(self, config):
        # filters_path
        self.filters_path = config.get("DEFAULT", "filters_path").split(",")

        # exec_dirs
        if config.has_option("DEFAULT", "exec_dirs"):
            self.exec_dirs = config.get("DEFAULT", "exec_dirs").split(",")
        else:
            # Use system PATH if exec_dirs is not specified
            self.exec_dirs = os.environ["PATH"].split(':')

        # syslog_log_facility
        if config.has_option("DEFAULT", "syslog_log_facility"):
            v = config.get("DEFAULT", "syslog_log_facility")
            facility_names = logging.handlers.SysLogHandler.facility_names
            self.syslog_log_facility = getattr(logging.handlers.SysLogHandler,
                                               v, None)
            if self.syslog_log_facility is None and v in facility_names:
                self.syslog_log_facility = facility_names.get(v)
            if self.syslog_log_facility is None:
                raise ValueError('Unexpected syslog_log_facility: %s' % v)
        else:
            default_facility = logging.handlers.SysLogHandler.LOG_SYSLOG
            self.syslog_log_facility = default_facility

        # syslog_log_level
        if config.has_option("DEFAULT", "syslog_log_level"):
            v = config.get("DEFAULT", "syslog_log_level")
            self.syslog_log_level = logging.getLevelName(v.upper())
            if (self.syslog_log_level == "Level %s" % v.upper()):
                raise ValueError('Unexepected syslog_log_level: %s' % v)
        else:
            self.syslog_log_level = logging.ERROR

        # use_syslog
        if config.has_option("DEFAULT", "use_syslog"):
            self.use_syslog = config.getboolean("DEFAULT", "use_syslog")
        else:
            self.use_syslog = False

def setup_syslog(execname, facility, level):
    rootwrap_logger = logging.getLogger()
    rootwrap_logger.setLevel(level)
    handler = logging.handlers.SysLogHandler(address='/dev/log',
                                             facility=facility)
    handler.setFormatter(logging.Formatter(
                         os.path.basename(execname) + ': %(message)s'))
    rootwrap_logger.addHandler(handler)

def build_filter(class_name, *args):
    """Returns a filter object of class class_name"""
    if not hasattr(filters, class_name):
        logging.warning("Skipping unknown filter class (%s) specified "
                        "in filter definitions" % class_name)
        return None
    filterclass = getattr(filters, class_name)
    return filterclass(*args)

def load_filters(filters_path):
    """Load filters from a list of directories"""
    filterlist = []
    for filterdir in filters_path:
        if not os.path.isdir(filterdir):
            continue
        for filterfile in os.listdir(filterdir):
            filterconfig = ConfigParser.RawConfigParser()
            filterconfig.read(os.path.join(filterdir, filterfile))
            for (name, value) in filterconfig.items("Filters"):
                filterdefinition = [string.strip(s) for s in value.split(',')]
                newfilter = build_filter(*filterdefinition)
                if newfilter is None:
                    continue
                newfilter.name = name
                filterlist.append(newfilter)
    return filterlist

def match_filter(filters, userargs, exec_dirs=[]):
    """
    Checks user command and arguments through command filters and
    returns the first matching filter.
    Raises NoFilterMatched if no filter matched.
    Raises FilterMatchNotExecutable if no executable was found for the
    best filter match.
    """
    first_not_executable_filter = None

    for f in filters:
        if f.match(userargs):
            # Try other filters if executable is absent
            if not f.get_exec(exec_dirs=exec_dirs):
                if not first_not_executable_filter:
                    first_not_executable_filter = f
                continue
            # Otherwise return matching filter for execution
            return f

    if first_not_executable_filter:
        # A filter matched, but no executable was found for it
        raise FilterMatchNotExecutable(match=first_not_executable_filter)

    # No filter matched
    raise NoFilterMatched()



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2014 Intel Inc.
# All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the"License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#  http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""The rgws api."""


import platform

from vsm.api.openstack import wsgi
from vsm.api.views import mdses as mds_views
from vsm import conductor
from vsm import exception
from vsm import flags
from vsm.openstack.common import log as logging
from vsm import scheduler
from vsm import utils

LOG = logging.getLogger(__name__)

FLAGS = flags.FLAGS


class RgwController(wsgi.Controller):
    """The rgw API controller for the VSM API."""
    _view_builder_class = mds_views.ViewBuilder

    def __init__(self, ext_mgr):
        self.conductor_api = conductor.API()
        self.scheduler_api = scheduler.API()
        self.ext_mgr = ext_mgr
        super(RgwController, self).__init__()

    def create(self, req, body):
        """
        Create a rgw.
        :param req:
        :param body:
        {
            "rgw": {
                "rgw_info": {
                    "server_name": "rgw-node1",
                    "rgw_instance_name": "radosgw.gateway",
                    "is_ssl": False
                },
                "user_info": {
                    "uid": "johndoe",
                    "display_name": "John Doe",
                    "email": "john@example.comjohn@example.com",
                    "sub_user": "johndoe:swift",
                    "access": "full",
                    "key_type": "swift"
                }
            }
        }
        :return:
        """
        context = req.environ['vsm.context']

        # check if the server has been installed radosgw

        def _get_os():
            (distro, release, codename) = platform.dist()
            return distro

        distro = _get_os()
        if distro.lower() == "centos":
            try:
                utils.execute("ls", "/usr/bin/radosgw", run_as_root=True)
            except:
                LOG.error("Not installed ceph-radosgw")
                raise exception.VsmException()
        elif distro.lower() == "ubuntu":
            try:
                utils.execute("ls", "/usr/bin/radosgw", run_as_root=True)
            except:
                LOG.error("Not installed radosgw")
                raise exception.VsmException()

        rgw = body['rgw']
        rgw_info = rgw['rgw_info']
        user_info = rgw['user_info']
        server_name = rgw_info.get('server_name', '')
        if not server_name:
            LOG.error("No server to create rgw.")
            raise exception.VsmException()
        rgw_instance_name = rgw_info.get('rgw_instance_name', '')
        if not rgw_instance_name:
            rgw_instance_name = "gateway"
            LOG.warn("rgw instance name uses default name 'rgw'")
        is_ssl = rgw_info.get('is_ssl', '')
        if is_ssl == '':
            is_ssl = False
            LOG.warn("do not use ssl")
        uid = user_info.get('uid') if user_info.get('uid', '') else "johndoe"
        display_name = user_info.get('display_name') if user_info.get('display_name', '') else "John Doe"
        email = user_info.get('email') if user_info.get('email', '') else "john@example.comjohn@example.com"
        sub_user = user_info.get('sub_user') if user_info.get('sub_user', '') else "johndoe:swift"
        access = user_info.get('access') if user_info.get('access', '') else "full"
        key_type = user_info.get('key_type') if user_info.get('key_type', '') else "swift"

        self.scheduler_api.rgw_create(context, server_name, rgw_instance_name, is_ssl,
                                      uid, display_name, email, sub_user, access, key_type)


def create_resource(ext_mgr):
    return wsgi.Resource(RgwController(ext_mgr))



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2011 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os.path

from lxml import etree

from vsm import utils

XMLNS_V10 = 'http://docs.rackspacecloud.com/servers/api/v1.0'
XMLNS_V11 = 'http://docs.openstack.org/compute/api/v1.1'
XMLNS_COMMON_V10 = 'http://docs.openstack.org/common/api/v1.0'
XMLNS_ATOM = 'http://www.w3.org/2005/Atom'
XMLNS_VOLUME_V1 = 'http://docs.openstack.org/storage/api/v1'
XMLNS_VOLUME_V2 = ('http://docs.openstack.org/api/openstack-storage/2.0/'
                   'content')

def validate_schema(xml, schema_name):
    if isinstance(xml, str):
        xml = etree.fromstring(xml)
    base_path = 'vsm/api/schemas/v1.1/'
    if schema_name in ('atom', 'atom-link'):
        base_path = 'vsm/api/schemas/'
    schema_path = os.path.join(utils.vsmdir(),
                               '%s%s.rng' % (base_path, schema_name))
    schema_doc = etree.parse(schema_path)
    relaxng = etree.RelaxNG(schema_doc)
    relaxng.assertValid(xml)

class Selector(object):
    """Selects datum to operate on from an object."""

    def __init__(self, *chain):
        """Initialize the selector.

        Each argument is a subsequent index into the object.
        """

        self.chain = chain

    def __repr__(self):
        """Return a representation of the selector."""

        return "Selector" + repr(self.chain)

    def __call__(self, obj, do_raise=False):
        """Select a datum to operate on.

        Selects the relevant datum within the object.

        :param obj: The object from which to select the object.
        :param do_raise: If False (the default), return None if the
                         indexed datum does not exist.  Otherwise,
                         raise a KeyError.
        """

        # Walk the selector list
        for elem in self.chain:
            # If it's callable, call it
            if callable(elem):
                obj = elem(obj)
            else:
                # Use indexing
                try:
                    obj = obj[elem]
                except (KeyError, IndexError):
                    # No sense going any further
                    if do_raise:
                        # Convert to a KeyError, for consistency
                        raise KeyError(elem)
                    return None

        # Return the finally-selected object
        return obj

def get_items(obj):
    """Get items in obj."""

    return list(obj.items())

class EmptyStringSelector(Selector):
    """Returns the empty string if Selector would return None."""
    def __call__(self, obj, do_raise=False):
        """Returns empty string if the selected value does not exist."""

        try:
            return super(EmptyStringSelector, self).__call__(obj, True)
        except KeyError:
            return ""

class ConstantSelector(object):
    """Returns a constant."""

    def __init__(self, value):
        """Initialize the selector.

        :param value: The value to return.
        """

        self.value = value

    def __repr__(self):
        """Return a representation of the selector."""

        return repr(self.value)

    def __call__(self, _obj, _do_raise=False):
        """Select a datum to operate on.

        Returns a constant value.  Compatible with
        Selector.__call__().
        """

        return self.value

class TemplateElement(object):
    """Represent an element in the template."""

    def __init__(self, tag, attrib=None, selector=None, subselector=None,
                 **extra):
        """Initialize an element.

        Initializes an element in the template.  Keyword arguments
        specify attributes to be set on the element; values must be
        callables.  See TemplateElement.set() for more information.

        :param tag: The name of the tag to create.
        :param attrib: An optional dictionary of element attributes.
        :param selector: An optional callable taking an object and
                         optional boolean do_raise indicator and
                         returning the object bound to the element.
        :param subselector: An optional callable taking an object and
                            optional boolean do_raise indicator and
                            returning the object bound to the element.
                            This is used to further refine the datum
                            object returned by selector in the event
                            that it is a list of objects.
        """

        # Convert selector into a Selector
        if selector is None:
            selector = Selector()
        elif not callable(selector):
            selector = Selector(selector)

        # Convert subselector into a Selector
        if subselector is not None and not callable(subselector):
            subselector = Selector(subselector)

        self.tag = tag
        self.selector = selector
        self.subselector = subselector
        self.attrib = {}
        self._text = None
        self._children = []
        self._childmap = {}

        # Run the incoming attributes through set() so that they
        # become selectorized
        if not attrib:
            attrib = {}
        attrib.update(extra)
        for k, v in attrib.items():
            self.set(k, v)

    def __repr__(self):
        """Return a representation of the template element."""

        return ('<%s.%s %r at %#x>' %
                (self.__class__.__module__, self.__class__.__name__,
                 self.tag, id(self)))

    def __len__(self):
        """Return the number of child elements."""

        return len(self._children)

    def __contains__(self, key):
        """Determine whether a child node named by key exists."""

        return key in self._childmap

    def __getitem__(self, idx):
        """Retrieve a child node by index or name."""

        if isinstance(idx, basestring):
            # Allow access by node name
            return self._childmap[idx]
        else:
            return self._children[idx]

    def append(self, elem):
        """Append a child to the element."""

        # Unwrap templates...
        elem = elem.unwrap()

        # Avoid duplications
        if elem.tag in self._childmap:
            raise KeyError(elem.tag)

        self._children.append(elem)
        self._childmap[elem.tag] = elem

    def extend(self, elems):
        """Append children to the element."""

        # Pre-evaluate the elements
        elemmap = {}
        elemlist = []
        for elem in elems:
            # Unwrap templates...
            elem = elem.unwrap()

            # Avoid duplications
            if elem.tag in self._childmap or elem.tag in elemmap:
                raise KeyError(elem.tag)

            elemmap[elem.tag] = elem
            elemlist.append(elem)

        # Update the children
        self._children.extend(elemlist)
        self._childmap.update(elemmap)

    def insert(self, idx, elem):
        """Insert a child element at the given index."""

        # Unwrap templates...
        elem = elem.unwrap()

        # Avoid duplications
        if elem.tag in self._childmap:
            raise KeyError(elem.tag)

        self._children.insert(idx, elem)
        self._childmap[elem.tag] = elem

    def remove(self, elem):
        """Remove a child element."""

        # Unwrap templates...
        elem = elem.unwrap()

        # Check if element exists
        if elem.tag not in self._childmap or self._childmap[elem.tag] != elem:
            raise ValueError(_('element is not a child'))

        self._children.remove(elem)
        del self._childmap[elem.tag]

    def get(self, key):
        """Get an attribute.

        Returns a callable which performs datum selection.

        :param key: The name of the attribute to get.
        """

        return self.attrib[key]

    def set(self, key, value=None):
        """Set an attribute.

        :param key: The name of the attribute to set.

        :param value: A callable taking an object and optional boolean
                      do_raise indicator and returning the datum bound
                      to the attribute.  If None, a Selector() will be
                      constructed from the key.  If a string, a
                      Selector() will be constructed from the string.
        """

        # Convert value to a selector
        if value is None:
            value = Selector(key)
        elif not callable(value):
            value = Selector(value)

        self.attrib[key] = value

    def keys(self):
        """Return the attribute names."""

        return self.attrib.keys()

    def items(self):
        """Return the attribute names and values."""

        return self.attrib.items()

    def unwrap(self):
        """Unwraps a template to return a template element."""

        # We are a template element
        return self

    def wrap(self):
        """Wraps a template element to return a template."""

        # Wrap in a basic Template
        return Template(self)

    def apply(self, elem, obj):
        """Apply text and attributes to an etree.Element.

        Applies the text and attribute instructions in the template
        element to an etree.Element instance.

        :param elem: An etree.Element instance.
        :param obj: The base object associated with this template
                    element.
        """

        # Start with the text...
        if self.text is not None:
            elem.text = unicode(self.text(obj))

        # Now set up all the attributes...
        for key, value in self.attrib.items():
            try:
                elem.set(key, unicode(value(obj, True)))
            except KeyError:
                # Attribute has no value, so don't include it
                pass

    def _render(self, parent, datum, patches, nsmap):
        """Internal rendering.

        Renders the template node into an etree.Element object.
        Returns the etree.Element object.

        :param parent: The parent etree.Element instance.
        :param datum: The datum associated with this template element.
        :param patches: A list of other template elements that must
                        also be applied.
        :param nsmap: An optional namespace dictionary to be
                      associated with the etree.Element instance.
        """

        # Allocate a node
        if callable(self.tag):
            tagname = self.tag(datum)
        else:
            tagname = self.tag
        elem = etree.Element(tagname, nsmap=nsmap)

        # If we have a parent, append the node to the parent
        if parent is not None:
            parent.append(elem)

        # If the datum is None, do nothing else
        if datum is None:
            return elem

        # Apply this template element to the element
        self.apply(elem, datum)

        # Additionally, apply the patches
        for patch in patches:
            patch.apply(elem, datum)

        # We have fully rendered the element; return it
        return elem

    def render(self, parent, obj, patches=[], nsmap=None):
        """Render an object.

        Renders an object against this template node.  Returns a list
        of two-item tuples, where the first item is an etree.Element
        instance and the second item is the datum associated with that
        instance.

        :param parent: The parent for the etree.Element instances.
        :param obj: The object to render this template element
                    against.
        :param patches: A list of other template elements to apply
                        when rendering this template element.
        :param nsmap: An optional namespace dictionary to attach to
                      the etree.Element instances.
        """

        # First, get the datum we're rendering
        data = None if obj is None else self.selector(obj)

        # Check if we should render at all
        if not self.will_render(data):
            return []
        elif data is None:
            return [(self._render(parent, None, patches, nsmap), None)]

        # Make the data into a list if it isn't already
        if not isinstance(data, list):
            data = [data]
        elif parent is None:
            raise ValueError(_('root element selecting a list'))

        # Render all the elements
        elems = []
        for datum in data:
            if self.subselector is not None:
                datum = self.subselector(datum)
            elems.append((self._render(parent, datum, patches, nsmap), datum))

        # Return all the elements rendered, as well as the
        # corresponding datum for the next step down the tree
        return elems

    def will_render(self, datum):
        """Hook method.

        An overridable hook method to determine whether this template
        element will be rendered at all.  By default, returns False
        (inhibiting rendering) if the datum is None.

        :param datum: The datum associated with this template element.
        """

        # Don't render if datum is None
        return datum is not None

    def _text_get(self):
        """Template element text.

        Either None or a callable taking an object and optional
        boolean do_raise indicator and returning the datum bound to
        the text of the template element.
        """

        return self._text

    def _text_set(self, value):
        # Convert value to a selector
        if value is not None and not callable(value):
            value = Selector(value)

        self._text = value

    def _text_del(self):
        self._text = None

    text = property(_text_get, _text_set, _text_del)

    def tree(self):
        """Return string representation of the template tree.

        Returns a representation of the template rooted at this
        element as a string, suitable for inclusion in debug logs.
        """

        # Build the inner contents of the tag...
        contents = [self.tag, '!selector=%r' % self.selector]

        # Add the text...
        if self.text is not None:
            contents.append('!text=%r' % self.text)

        # Add all the other attributes
        for key, value in self.attrib.items():
            contents.append('%s=%r' % (key, value))

        # If there are no children, return it as a closed tag
        if len(self) == 0:
            return '<%s/>' % ' '.join([str(i) for i in contents])

        # OK, recurse to our children
        children = [c.tree() for c in self]

        # Return the result
        return ('<%s>%s</%s>' %
                (' '.join(contents), ''.join(children), self.tag))

def SubTemplateElement(parent, tag, attrib=None, selector=None,
                       subselector=None, **extra):
    """Create a template element as a child of another.

    Corresponds to the etree.SubElement interface.  Parameters are as
    for TemplateElement, with the addition of the parent.
    """

    # Convert attributes
    attrib = attrib or {}
    attrib.update(extra)

    # Get a TemplateElement
    elem = TemplateElement(tag, attrib=attrib, selector=selector,
                           subselector=subselector)

    # Append the parent safely
    if parent is not None:
        parent.append(elem)

    return elem

class Template(object):
    """Represent a template."""

    def __init__(self, root, nsmap=None):
        """Initialize a template.

        :param root: The root element of the template.
        :param nsmap: An optional namespace dictionary to be
                      associated with the root element of the
                      template.
        """

        self.root = root.unwrap() if root is not None else None
        self.nsmap = nsmap or {}
        self.serialize_options = dict(encoding='UTF-8', xml_declaration=True)

    def _serialize(self, parent, obj, siblings, nsmap=None):
        """Internal serialization.

        Recursive routine to build a tree of etree.Element instances
        from an object based on the template.  Returns the first
        etree.Element instance rendered, or None.

        :param parent: The parent etree.Element instance.  Can be
                       None.
        :param obj: The object to render.
        :param siblings: The TemplateElement instances against which
                         to render the object.
        :param nsmap: An optional namespace dictionary to be
                      associated with the etree.Element instance
                      rendered.
        """

        # First step, render the element
        elems = siblings[0].render(parent, obj, siblings[1:], nsmap)

        # Now, recurse to all child elements
        seen = set()
        for idx, sibling in enumerate(siblings):
            for child in sibling:
                # Have we handled this child already?
                if child.tag in seen:
                    continue
                seen.add(child.tag)

                # Determine the child's siblings
                nieces = [child]
                for sib in siblings[idx + 1:]:
                    if child.tag in sib:
                        nieces.append(sib[child.tag])

                # Now we recurse for every data element
                for elem, datum in elems:
                    self._serialize(elem, datum, nieces)

        # Return the first element; at the top level, this will be the
        # root element
        if elems:
            return elems[0][0]

    def serialize(self, obj, *args, **kwargs):
        """Serialize an object.

        Serializes an object against the template.  Returns a string
        with the serialized XML.  Positional and keyword arguments are
        passed to etree.tostring().

        :param obj: The object to serialize.
        """

        elem = self.make_tree(obj)
        if elem is None:
            return ''

        for k, v in self.serialize_options.items():
            kwargs.setdefault(k, v)

        # Serialize it into XML
        return etree.tostring(elem, *args, **kwargs)

    def make_tree(self, obj):
        """Create a tree.

        Serializes an object against the template.  Returns an Element
        node with appropriate children.

        :param obj: The object to serialize.
        """

        # If the template is empty, return the empty string
        if self.root is None:
            return None

        # Get the siblings and nsmap of the root element
        siblings = self._siblings()
        nsmap = self._nsmap()

        # Form the element tree
        return self._serialize(None, obj, siblings, nsmap)

    def _siblings(self):
        """Hook method for computing root siblings.

        An overridable hook method to return the siblings of the root
        element.  By default, this is the root element itself.
        """

        return [self.root]

    def _nsmap(self):
        """Hook method for computing the namespace dictionary.

        An overridable hook method to return the namespace dictionary.
        """

        return self.nsmap.copy()

    def unwrap(self):
        """Unwraps a template to return a template element."""

        # Return the root element
        return self.root

    def wrap(self):
        """Wraps a template element to return a template."""

        # We are a template
        return self

    def apply(self, master):
        """Hook method for determining slave applicability.

        An overridable hook method used to determine if this template
        is applicable as a slave to a given master template.

        :param master: The master template to test.
        """

        return True

    def tree(self):
        """Return string representation of the template tree.

        Returns a representation of the template as a string, suitable
        for inclusion in debug logs.
        """

        return "%r: %s" % (self, self.root.tree())

class MasterTemplate(Template):
    """Represent a master template.

    Master templates are versioned derivatives of templates that
    additionally allow slave templates to be attached.  Slave
    templates allow modification of the serialized result without
    directly changing the master.
    """

    def __init__(self, root, version, nsmap=None):
        """Initialize a master template.

        :param root: The root element of the template.
        :param version: The version number of the template.
        :param nsmap: An optional namespace dictionary to be
                      associated with the root element of the
                      template.
        """

        super(MasterTemplate, self).__init__(root, nsmap)
        self.version = version
        self.slaves = []

    def __repr__(self):
        """Return string representation of the template."""

        return ("<%s.%s object version %s at %#x>" %
                (self.__class__.__module__, self.__class__.__name__,
                 self.version, id(self)))

    def _siblings(self):
        """Hook method for computing root siblings.

        An overridable hook method to return the siblings of the root
        element.  This is the root element plus the root elements of
        all the slave templates.
        """

        return [self.root] + [slave.root for slave in self.slaves]

    def _nsmap(self):
        """Hook method for computing the namespace dictionary.

        An overridable hook method to return the namespace dictionary.
        The namespace dictionary is computed by taking the master
        template's namespace dictionary and updating it from all the
        slave templates.
        """

        nsmap = self.nsmap.copy()
        for slave in self.slaves:
            nsmap.update(slave._nsmap())
        return nsmap

    def attach(self, *slaves):
        """Attach one or more slave templates.

        Attaches one or more slave templates to the master template.
        Slave templates must have a root element with the same tag as
        the master template.  The slave template's apply() method will
        be called to determine if the slave should be applied to this
        master; if it returns False, that slave will be skipped.
        (This allows filtering of slaves based on the version of the
        master template.)
        """

        slave_list = []
        for slave in slaves:
            slave = slave.wrap()

            # Make sure we have a tree match
            if slave.root.tag != self.root.tag:
                slavetag = slave.root.tag
                mastertag = self.root.tag
                msg = _("Template tree mismatch; adding slave %(slavetag)s "
                        "to master %(mastertag)s") % locals()
                raise ValueError(msg)

            # Make sure slave applies to this template
            if not slave.apply(self):
                continue

            slave_list.append(slave)

        # Add the slaves
        self.slaves.extend(slave_list)

    def copy(self):
        """Return a copy of this master template."""

        # Return a copy of the MasterTemplate
        tmp = self.__class__(self.root, self.version, self.nsmap)
        tmp.slaves = self.slaves[:]
        return tmp

class SlaveTemplate(Template):
    """Represent a slave template.

    Slave templates are versioned derivatives of templates.  Each
    slave has a minimum version and optional maximum version of the
    master template to which they can be attached.
    """

    def __init__(self, root, min_vers, max_vers=None, nsmap=None):
        """Initialize a slave template.

        :param root: The root element of the template.
        :param min_vers: The minimum permissible version of the master
                         template for this slave template to apply.
        :param max_vers: An optional upper bound for the master
                         template version.
        :param nsmap: An optional namespace dictionary to be
                      associated with the root element of the
                      template.
        """

        super(SlaveTemplate, self).__init__(root, nsmap)
        self.min_vers = min_vers
        self.max_vers = max_vers

    def __repr__(self):
        """Return string representation of the template."""

        return ("<%s.%s object versions %s-%s at %#x>" %
                (self.__class__.__module__, self.__class__.__name__,
                 self.min_vers, self.max_vers, id(self)))

    def apply(self, master):
        """Hook method for determining slave applicability.

        An overridable hook method used to determine if this template
        is applicable as a slave to a given master template.  This
        version requires the master template to have a version number
        between min_vers and max_vers.

        :param master: The master template to test.
        """

        # Does the master meet our minimum version requirement?
        if master.version < self.min_vers:
            return False

        # How about our maximum version requirement?
        if self.max_vers is not None and master.version > self.max_vers:
            return False

        return True

class TemplateBuilder(object):
    """Template builder.

    This class exists to allow templates to be lazily built without
    having to build them each time they are needed.  It must be
    subclassed, and the subclass must implement the construct()
    method, which must return a Template (or subclass) instance.  The
    constructor will always return the template returned by
    construct(), or, if it has a copy() method, a copy of that
    template.
    """

    _tmpl = None

    def __new__(cls, copy=True):
        """Construct and return a template.

        :param copy: If True (the default), a copy of the template
                     will be constructed and returned, if possible.
        """

        # Do we need to construct the template?
        if cls._tmpl is None:
            tmp = super(TemplateBuilder, cls).__new__(cls)

            # Construct the template
            cls._tmpl = tmp.construct()

        # If the template has a copy attribute, return the result of
        # calling it
        if copy and hasattr(cls._tmpl, 'copy'):
            return cls._tmpl.copy()

        # Return the template
        return cls._tmpl

    def construct(self):
        """Construct a template.

        Called to construct a template instance, which it must return.
        Only called once.
        """

        raise NotImplementedError(_("subclasses must implement construct()!"))

def make_links(parent, selector=None):
    """
    Attach an Atom <links> element to the parent.
    """

    elem = SubTemplateElement(parent, '{%s}link' % XMLNS_ATOM,
                              selector=selector)
    elem.set('rel')
    elem.set('type')
    elem.set('href')

    # Just for completeness...
    return elem

def make_flat_dict(name, selector=None, subselector=None, ns=None):
    """
    Utility for simple XML templates that traditionally used
    XMLDictSerializer with no metadata.  Returns a template element
    where the top-level element has the given tag name, and where
    sub-elements have tag names derived from the object's keys and
    text derived from the object's values.  This only works for flat
    dictionary objects, not dictionaries containing nested lists or
    dictionaries.
    """

    # Set up the names we need...
    if ns is None:
        elemname = name
        tagname = Selector(0)
    else:
        elemname = '{%s}%s' % (ns, name)
        tagname = lambda obj, do_raise=False: '{%s}%s' % (ns, obj[0])

    if selector is None:
        selector = name

    # Build the root element
    root = TemplateElement(elemname, selector=selector,
                           subselector=subselector)

    # Build an element to represent all the keys and values
    elem = SubTemplateElement(root, tagname, selector=get_items)
    elem.text = 1

    # Return the template
    return root



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
# Copyright 2011 - 2012, Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Shared code between AMQP based openstack.common.rpc implementations.

The code in this module is shared between the rpc implemenations based on AMQP.
Specifically, this includes impl_kombu and impl_qpid.  impl_carrot also uses
AMQP, but is deprecated and predates this code.
"""

import collections
import inspect
import sys
import uuid

from eventlet import greenpool
from eventlet import pools
from eventlet import queue
from eventlet import semaphore
# TODO(pekowsk): Remove import cfg and below comment in Havana.
# This import should no longer be needed when the amqp_rpc_single_reply_queue
# option is removed.
from oslo.config import cfg

from vsm.openstack.common import excutils
from vsm.openstack.common.gettextutils import _
from vsm.openstack.common import local
from vsm.openstack.common import log as logging
from vsm.openstack.common.rpc import common as rpc_common

# TODO(pekowski): Remove this option in Havana.
amqp_opts = [
    cfg.BoolOpt('amqp_rpc_single_reply_queue',
                default=False,
                help='Enable a fast single reply queue if using AMQP based '
                'RPC like RabbitMQ or Qpid.'),
]

cfg.CONF.register_opts(amqp_opts)

UNIQUE_ID = '_unique_id'
LOG = logging.getLogger(__name__)

class Pool(pools.Pool):
    """Class that implements a Pool of Connections."""
    def __init__(self, conf, connection_cls, *args, **kwargs):
        self.connection_cls = connection_cls
        self.conf = conf
        kwargs.setdefault("max_size", self.conf.rpc_conn_pool_size)
        kwargs.setdefault("order_as_stack", True)
        super(Pool, self).__init__(*args, **kwargs)
        self.reply_proxy = None

    # TODO(comstud): Timeout connections not used in a while
    def create(self):
        LOG.debug(_('Pool creating new connection'))
        return self.connection_cls(self.conf)

    def empty(self):
        while self.free_items:
            self.get().close()
        # Force a new connection pool to be created.
        # Note that this was added due to failing unit test cases. The issue
        # is the above "while loop" gets all the cached connections from the
        # pool and closes them, but never returns them to the pool, a pool
        # leak. The unit tests hang waiting for an item to be returned to the
        # pool. The unit tests get here via the teatDown() method. In the run
        # time code, it gets here via cleanup() and only appears in service.py
        # just before doing a sys.exit(), so cleanup() only happens once and
        # the leakage is not a problem.
        self.connection_cls.pool = None

_pool_create_sem = semaphore.Semaphore()

def get_connection_pool(conf, connection_cls):
    with _pool_create_sem:
        # Make sure only one thread tries to create the connection pool.
        if not connection_cls.pool:
            connection_cls.pool = Pool(conf, connection_cls)
    return connection_cls.pool

class ConnectionContext(rpc_common.Connection):
    """The class that is actually returned to the caller of
    create_connection().  This is essentially a wrapper around
    Connection that supports 'with'.  It can also return a new
    Connection, or one from a pool.  The function will also catch
    when an instance of this class is to be deleted.  With that
    we can return Connections to the pool on exceptions and so
    forth without making the caller be responsible for catching
    them.  If possible the function makes sure to return a
    connection to the pool.
    """

    def __init__(self, conf, connection_pool, pooled=True, server_params=None):
        """Create a new connection, or get one from the pool"""
        self.connection = None
        self.conf = conf
        self.connection_pool = connection_pool
        if pooled:
            self.connection = connection_pool.get()
        else:
            self.connection = connection_pool.connection_cls(
                conf,
                server_params=server_params)
        self.pooled = pooled

    def __enter__(self):
        """When with ConnectionContext() is used, return self"""
        return self

    def _done(self):
        """If the connection came from a pool, clean it up and put it back.
        If it did not come from a pool, close it.
        """
        if self.connection:
            if self.pooled:
                # Reset the connection so it's ready for the next caller
                # to grab from the pool
                self.connection.reset()
                self.connection_pool.put(self.connection)
            else:
                try:
                    self.connection.close()
                except Exception:
                    pass
            self.connection = None

    def __exit__(self, exc_type, exc_value, tb):
        """End of 'with' statement.  We're done here."""
        self._done()

    def __del__(self):
        """Caller is done with this connection.  Make sure we cleaned up."""
        self._done()

    def close(self):
        """Caller is done with this connection."""
        self._done()

    def create_consumer(self, topic, proxy, fanout=False):
        self.connection.create_consumer(topic, proxy, fanout)

    def create_worker(self, topic, proxy, pool_name):
        self.connection.create_worker(topic, proxy, pool_name)

    def join_consumer_pool(self, callback, pool_name, topic, exchange_name):
        self.connection.join_consumer_pool(callback,
                                           pool_name,
                                           topic,
                                           exchange_name)

    def consume_in_thread(self):
        self.connection.consume_in_thread()

    def __getattr__(self, key):
        """Proxy all other calls to the Connection instance"""
        if self.connection:
            return getattr(self.connection, key)
        else:
            raise rpc_common.InvalidRPCConnectionReuse()

class ReplyProxy(ConnectionContext):
    """ Connection class for RPC replies / callbacks """
    def __init__(self, conf, connection_pool):
        self._call_waiters = {}
        self._num_call_waiters = 0
        self._num_call_waiters_wrn_threshhold = 10
        self._reply_q = 'reply_' + uuid.uuid4().hex
        super(ReplyProxy, self).__init__(conf, connection_pool, pooled=False)
        self.declare_direct_consumer(self._reply_q, self._process_data)
        self.consume_in_thread()

    def _process_data(self, message_data):
        msg_id = message_data.pop('_msg_id', None)
        waiter = self._call_waiters.get(msg_id)
        if not waiter:
            LOG.warn(_('no calling threads waiting for msg_id : %s'
                       ', message : %s') % (msg_id, message_data))
        else:
            waiter.put(message_data)

    def add_call_waiter(self, waiter, msg_id):
        self._num_call_waiters += 1
        if self._num_call_waiters > self._num_call_waiters_wrn_threshhold:
            LOG.warn(_('Number of call waiters is greater than warning '
                       'threshhold: %d. There could be a MulticallProxyWaiter '
                       'leak.') % self._num_call_waiters_wrn_threshhold)
            self._num_call_waiters_wrn_threshhold *= 2
        self._call_waiters[msg_id] = waiter

    def del_call_waiter(self, msg_id):
        self._num_call_waiters -= 1
        del self._call_waiters[msg_id]

    def get_reply_q(self):
        return self._reply_q

def msg_reply(conf, msg_id, reply_q, connection_pool, reply=None,
              failure=None, ending=False, log_failure=True):
    """Sends a reply or an error on the channel signified by msg_id.

    Failure should be a sys.exc_info() tuple.

    """
    with ConnectionContext(conf, connection_pool) as conn:
        if failure:
            failure = rpc_common.serialize_remote_exception(failure,
                                                            log_failure)

        try:
            msg = {'result': reply, 'failure': failure}
        except TypeError:
            msg = {'result': dict((k, repr(v))
                   for k, v in reply.__dict__.iteritems()),
                   'failure': failure}
        if ending:
            msg['ending'] = True
        _add_unique_id(msg)
        # If a reply_q exists, add the msg_id to the reply and pass the
        # reply_q to direct_send() to use it as the response queue.
        # Otherwise use the msg_id for backward compatibilty.
        if reply_q:
            msg['_msg_id'] = msg_id
            conn.direct_send(reply_q, rpc_common.serialize_msg(msg))
        else:
            conn.direct_send(msg_id, rpc_common.serialize_msg(msg))

class RpcContext(rpc_common.CommonRpcContext):
    """Context that supports replying to a rpc.call"""
    def __init__(self, **kwargs):
        self.msg_id = kwargs.pop('msg_id', None)
        self.reply_q = kwargs.pop('reply_q', None)
        self.conf = kwargs.pop('conf')
        super(RpcContext, self).__init__(**kwargs)

    def deepcopy(self):
        values = self.to_dict()
        values['conf'] = self.conf
        values['msg_id'] = self.msg_id
        values['reply_q'] = self.reply_q
        return self.__class__(**values)

    def reply(self, reply=None, failure=None, ending=False,
              connection_pool=None, log_failure=True):
        if self.msg_id:
            msg_reply(self.conf, self.msg_id, self.reply_q, connection_pool,
                      reply, failure, ending, log_failure)
            if ending:
                self.msg_id = None

def unpack_context(conf, msg):
    """Unpack context from msg."""
    context_dict = {}
    for key in list(msg.keys()):
        # NOTE(vish): Some versions of python don't like unicode keys
        #             in kwargs.
        key = str(key)
        if key.startswith('_context_'):
            value = msg.pop(key)
            context_dict[key[9:]] = value
    context_dict['msg_id'] = msg.pop('_msg_id', None)
    context_dict['reply_q'] = msg.pop('_reply_q', None)
    context_dict['conf'] = conf
    ctx = RpcContext.from_dict(context_dict)
    rpc_common._safe_log(LOG.debug, _('unpacked context: %s'), ctx.to_dict())
    return ctx

def pack_context(msg, context):
    """Pack context into msg.

    Values for message keys need to be less than 255 chars, so we pull
    context out into a bunch of separate keys. If we want to support
    more arguments in rabbit messages, we may want to do the same
    for args at some point.

    """
    context_d = dict([('_context_%s' % key, value)
                      for (key, value) in context.to_dict().iteritems()])
    msg.update(context_d)

class _MsgIdCache(object):
    """This class checks any duplicate messages."""

    # NOTE: This value is considered can be a configuration item, but
    #       it is not necessary to change its value in most cases,
    #       so let this value as static for now.
    DUP_MSG_CHECK_SIZE = 16

    def __init__(self, **kwargs):
        self.prev_msgids = collections.deque([],
                                             maxlen=self.DUP_MSG_CHECK_SIZE)

    def check_duplicate_message(self, message_data):
        """AMQP consumers may read same message twice when exceptions occur
           before ack is returned. This method prevents doing it.
        """
        if UNIQUE_ID in message_data:
            msg_id = message_data[UNIQUE_ID]
            if msg_id not in self.prev_msgids:
                self.prev_msgids.append(msg_id)
            else:
                raise rpc_common.DuplicateMessageError(msg_id=msg_id)

def _add_unique_id(msg):
    """Add unique_id for checking duplicate messages."""
    unique_id = uuid.uuid4().hex
    msg.update({UNIQUE_ID: unique_id})
    #LOG.debug(_('UNIQUE_ID is %s.') % (unique_id))

class _ThreadPoolWithWait(object):
    """Base class for a delayed invocation manager used by
    the Connection class to start up green threads
    to handle incoming messages.
    """

    def __init__(self, conf, connection_pool):
        self.pool = greenpool.GreenPool(conf.rpc_thread_pool_size)
        self.connection_pool = connection_pool
        self.conf = conf

    def wait(self):
        """Wait for all callback threads to exit."""
        self.pool.waitall()

class CallbackWrapper(_ThreadPoolWithWait):
    """Wraps a straight callback to allow it to be invoked in a green
    thread.
    """

    def __init__(self, conf, callback, connection_pool):
        """
        :param conf: cfg.CONF instance
        :param callback: a callable (probably a function)
        :param connection_pool: connection pool as returned by
                                get_connection_pool()
        """
        super(CallbackWrapper, self).__init__(
            conf=conf,
            connection_pool=connection_pool,
        )
        self.callback = callback

    def __call__(self, message_data):
        self.pool.spawn_n(self.callback, message_data)

class ProxyCallback(_ThreadPoolWithWait):
    """Calls methods on a proxy object based on method and args."""

    def __init__(self, conf, proxy, connection_pool):
        super(ProxyCallback, self).__init__(
            conf=conf,
            connection_pool=connection_pool,
        )
        self.proxy = proxy
        self.msg_id_cache = _MsgIdCache()

    def __call__(self, message_data):
        """Consumer callback to call a method on a proxy object.

        Parses the message for validity and fires off a thread to call the
        proxy object method.

        Message data should be a dictionary with two keys:
            method: string representing the method to call
            args: dictionary of arg: value

        Example: {'method': 'echo', 'args': {'value': 42}}

        """
        # It is important to clear the context here, because at this point
        # the previous context is stored in local.store.context
        if hasattr(local.store, 'context'):
            del local.store.context
        rpc_common._safe_log(LOG.debug, _('received %s'), message_data)
        self.msg_id_cache.check_duplicate_message(message_data)
        ctxt = unpack_context(self.conf, message_data)
        method = message_data.get('method')
        args = message_data.get('args', {})
        version = message_data.get('version', None)
        if not method:
            LOG.warn(_('no method for message: %s') % message_data)
            ctxt.reply(_('No method for message: %s') % message_data,
                       connection_pool=self.connection_pool)
            return
        self.pool.spawn_n(self._process_data, ctxt, version, method, args)

    def _process_data(self, ctxt, version, method, args):
        """Process a message in a new thread.

        If the proxy object we have has a dispatch method
        (see rpc.dispatcher.RpcDispatcher), pass it the version,
        method, and args and let it dispatch as appropriate.  If not, use
        the old behavior of magically calling the specified method on the
        proxy we have here.
        """
        ctxt.update_store()
        try:
            rval = self.proxy.dispatch(ctxt, version, method, **args)
            # Check if the result was a generator
            if inspect.isgenerator(rval):
                for x in rval:
                    ctxt.reply(x, None, connection_pool=self.connection_pool)
            else:
                ctxt.reply(rval, None, connection_pool=self.connection_pool)
            # This final None tells multicall that it is done.
            ctxt.reply(ending=True, connection_pool=self.connection_pool)
        except rpc_common.ClientException as e:
            LOG.debug(_('Expected exception during message handling (%s)') %
                      e._exc_info[1])
            ctxt.reply(None, e._exc_info,
                       connection_pool=self.connection_pool,
                       log_failure=False)
        except Exception:
            # sys.exc_info() is deleted by LOG.exception().
            exc_info = sys.exc_info()
            LOG.error(_('Exception during message handling'),
                      exc_info=exc_info)
            ctxt.reply(None, exc_info, connection_pool=self.connection_pool)

class MulticallProxyWaiter(object):
    def __init__(self, conf, msg_id, timeout, connection_pool):
        self._msg_id = msg_id
        self._timeout = timeout or conf.rpc_response_timeout
        self._reply_proxy = connection_pool.reply_proxy
        self._done = False
        self._got_ending = False
        self._conf = conf
        self._dataqueue = queue.LightQueue()
        # Add this caller to the reply proxy's call_waiters
        self._reply_proxy.add_call_waiter(self, self._msg_id)
        self.msg_id_cache = _MsgIdCache()

    def put(self, data):
        self._dataqueue.put(data)

    def done(self):
        if self._done:
            return
        self._done = True
        # Remove this caller from reply proxy's call_waiters
        self._reply_proxy.del_call_waiter(self._msg_id)

    def _process_data(self, data):
        result = None
        self.msg_id_cache.check_duplicate_message(data)
        if data['failure']:
            failure = data['failure']
            result = rpc_common.deserialize_remote_exception(self._conf,
                                                             failure)
        elif data.get('ending', False):
            self._got_ending = True
        else:
            result = data['result']
        return result

    def __iter__(self):
        """Return a result until we get a reply with an 'ending" flag"""
        if self._done:
            raise StopIteration
        while True:
            try:
                data = self._dataqueue.get(timeout=self._timeout)
                result = self._process_data(data)
            except queue.Empty:
                LOG.exception(_('Timed out waiting for RPC response.'))
                self.done()
                raise rpc_common.Timeout()
            except Exception:
                with excutils.save_and_reraise_exception():
                    self.done()
            if self._got_ending:
                self.done()
                raise StopIteration
            if isinstance(result, Exception):
                self.done()
                raise result
            yield result

#TODO(pekowski): Remove MulticallWaiter() in Havana.
class MulticallWaiter(object):
    def __init__(self, conf, connection, timeout):
        self._connection = connection
        self._iterator = connection.iterconsume(timeout=timeout or
                                                conf.rpc_response_timeout)
        self._result = None
        self._done = False
        self._got_ending = False
        self._conf = conf
        self.msg_id_cache = _MsgIdCache()

    def done(self):
        if self._done:
            return
        self._done = True
        self._iterator.close()
        self._iterator = None
        self._connection.close()

    def __call__(self, data):
        """The consume() callback will call this.  Store the result."""
        self.msg_id_cache.check_duplicate_message(data)
        if data['failure']:
            failure = data['failure']
            self._result = rpc_common.deserialize_remote_exception(self._conf,
                                                                   failure)

        elif data.get('ending', False):
            self._got_ending = True
        else:
            self._result = data['result']

    def __iter__(self):
        """Return a result until we get a 'None' response from consumer"""
        if self._done:
            raise StopIteration
        while True:
            try:
                self._iterator.next()
            except Exception:
                with excutils.save_and_reraise_exception():
                    self.done()
            if self._got_ending:
                self.done()
                raise StopIteration
            result = self._result
            if isinstance(result, Exception):
                self.done()
                raise result
            yield result

def create_connection(conf, new, connection_pool):
    """Create a connection"""
    return ConnectionContext(conf, connection_pool, pooled=not new)

_reply_proxy_create_sem = semaphore.Semaphore()

def multicall(conf, context, topic, msg, timeout, connection_pool):
    """Make a call that returns multiple times."""
    # TODO(pekowski): Remove all these comments in Havana.
    # For amqp_rpc_single_reply_queue = False,
    # Can't use 'with' for multicall, as it returns an iterator
    # that will continue to use the connection.  When it's done,
    # connection.close() will get called which will put it back into
    # the pool
    # For amqp_rpc_single_reply_queue = True,
    # The 'with' statement is mandatory for closing the connection
    #LOG.debug(_('Making synchronous call on %s ...'), topic)
    msg_id = uuid.uuid4().hex
    msg.update({'_msg_id': msg_id})
    #LOG.debug(_('MSG_ID is %s') % (msg_id))
    _add_unique_id(msg)
    pack_context(msg, context)

    # TODO(pekowski): Remove this flag and the code under the if clause
    #                 in Havana.
    if not conf.amqp_rpc_single_reply_queue:
        conn = ConnectionContext(conf, connection_pool)
        wait_msg = MulticallWaiter(conf, conn, timeout)
        conn.declare_direct_consumer(msg_id, wait_msg)
        conn.topic_send(topic, rpc_common.serialize_msg(msg), timeout)
    else:
        with _reply_proxy_create_sem:
            if not connection_pool.reply_proxy:
                connection_pool.reply_proxy = ReplyProxy(conf, connection_pool)
        msg.update({'_reply_q': connection_pool.reply_proxy.get_reply_q()})
        wait_msg = MulticallProxyWaiter(conf, msg_id, timeout, connection_pool)
        with ConnectionContext(conf, connection_pool) as conn:
            conn.topic_send(topic, rpc_common.serialize_msg(msg), timeout)
    return wait_msg

def call(conf, context, topic, msg, timeout, connection_pool):
    """Sends a message on a topic and wait for a response."""
    rv = multicall(conf, context, topic, msg, timeout, connection_pool)
    # NOTE(vish): return the last result from the multicall
    rv = list(rv)
    if not rv:
        return
    return rv[-1]

def cast(conf, context, topic, msg, connection_pool):
    """Sends a message on a topic without waiting for a response."""
    LOG.debug(_('Making asynchronous cast on %s...'), topic)
    _add_unique_id(msg)
    pack_context(msg, context)
    with ConnectionContext(conf, connection_pool) as conn:
        conn.topic_send(topic, rpc_common.serialize_msg(msg))

def fanout_cast(conf, context, topic, msg, connection_pool):
    """Sends a message on a fanout exchange without waiting for a response."""
    LOG.debug(_('Making asynchronous fanout cast...'))
    _add_unique_id(msg)
    pack_context(msg, context)
    with ConnectionContext(conf, connection_pool) as conn:
        conn.fanout_send(topic, rpc_common.serialize_msg(msg))

def cast_to_server(conf, context, server_params, topic, msg, connection_pool):
    """Sends a message on a topic to a specific server."""
    _add_unique_id(msg)
    pack_context(msg, context)
    with ConnectionContext(conf, connection_pool, pooled=False,
                           server_params=server_params) as conn:
        conn.topic_send(topic, rpc_common.serialize_msg(msg))

def fanout_cast_to_server(conf, context, server_params, topic, msg,
                          connection_pool):
    """Sends a message on a fanout exchange to a specific server."""
    _add_unique_id(msg)
    pack_context(msg, context)
    with ConnectionContext(conf, connection_pool, pooled=False,
                           server_params=server_params) as conn:
        conn.fanout_send(topic, rpc_common.serialize_msg(msg))

def notify(conf, context, topic, msg, connection_pool, envelope):
    """Sends a notification event on a topic."""
    LOG.debug(_('Sending %(event_type)s on %(topic)s'),
              dict(event_type=msg.get('event_type'),
                   topic=topic))
    _add_unique_id(msg)
    pack_context(msg, context)
    with ConnectionContext(conf, connection_pool) as conn:
        if envelope:
            msg = rpc_common.serialize_msg(msg, force_envelope=True)
        conn.notify_send(topic, msg)

def cleanup(connection_pool):
    if connection_pool:
        connection_pool.empty()

def get_control_exchange(conf):
    return conf.control_exchange



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2014 Intel Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Common Data
# Please input your common data here.

# ceph version and code
CEPH_FIREFLY = "firefly"
CEPH_GIANT = "giant"
CEPH_HAMMER = "hammer"
CEPH_INFERNALIS = "infernalis"
CEPH_JEWEL = "jewel"
CEPH_VERSION_CODE = [("0.80", CEPH_FIREFLY),
                     ("0.87", CEPH_GIANT),
                     ("0.94", CEPH_HAMMER),
                     ("9.2", CEPH_INFERNALIS),
                     ("10.2", CEPH_JEWEL)]



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
# Copyright 2011 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
A remote procedure call (rpc) abstraction.

For some wrappers that add message versioning to rpc, see:
    rpc.dispatcher
    rpc.proxy
"""

import inspect
import logging

from oslo.config import cfg

from vsm.openstack.common.gettextutils import _
from vsm.openstack.common import importutils
from vsm.openstack.common import local
from vsm.openstack.common.rpc.common import Timeout

LOG = logging.getLogger(__name__)

rpc_opts = [
    cfg.StrOpt('rpc_backend',
               default='%s.impl_kombu' % __package__,
               help="The messaging module to use, defaults to kombu."),
    cfg.IntOpt('rpc_thread_pool_size',
               default=64,
               help='Size of RPC thread pool'),
    cfg.IntOpt('rpc_conn_pool_size',
               default=30,
               help='Size of RPC connection pool'),
    cfg.IntOpt('rpc_response_timeout',
               default=60,
               help='Seconds to wait for a response from call or multicall'),
    cfg.IntOpt('rpc_cast_timeout',
               default=30,
               help='Seconds to wait before a cast expires (TTL). '
                    'Only supported by impl_zmq.'),
    cfg.ListOpt('allowed_rpc_exception_modules',
                default=['vsm.openstack.common.exception',
                         'nova.exception',
                         'vsm.exception',
                         'exceptions',
                         ],
                help='Modules of exceptions that are permitted to be recreated'
                     'upon receiving exception data from an rpc call.'),
    cfg.BoolOpt('fake_rabbit',
                default=False,
                help='If passed, use a fake RabbitMQ provider'),
    cfg.StrOpt('control_exchange',
               default='openstack',
               help='AMQP exchange to connect to if using RabbitMQ or Qpid'),
]

CONF = cfg.CONF
CONF.register_opts(rpc_opts)

def set_defaults(control_exchange):
    cfg.set_defaults(rpc_opts,
                     control_exchange=control_exchange)

def create_connection(new=True):
    """Create a connection to the message bus used for rpc.

    For some example usage of creating a connection and some consumers on that
    connection, see nova.service.

    :param new: Whether or not to create a new connection.  A new connection
                will be created by default.  If new is False, the
                implementation is free to return an existing connection from a
                pool.

    :returns: An instance of openstack.common.rpc.common.Connection
    """
    return _get_impl().create_connection(CONF, new=new)

def _check_for_lock():
    if not CONF.debug:
        return None

    if ((hasattr(local.strong_store, 'locks_held')
         and local.strong_store.locks_held)):
        stack = ' :: '.join([frame[3] for frame in inspect.stack()])
        LOG.warn(_('A RPC is being made while holding a lock. The locks '
                   'currently held are %(locks)s. This is probably a bug. '
                   'Please report it. Include the following: [%(stack)s].'),
                 {'locks': local.strong_store.locks_held,
                  'stack': stack})
        return True

    return False

def __call(context, topic, msg, timeout=None, check_for_lock=False):
    """Invoke a remote method that returns something.

    :param context: Information that identifies the user that has made this
                    request.
    :param topic: The topic to send the rpc message to.  This correlates to the
                  topic argument of
                  openstack.common.rpc.common.Connection.create_consumer()
                  and only applies when the consumer was created with
                  fanout=False.
    :param msg: This is a dict in the form { "method" : "method_to_invoke",
                                             "args" : dict_of_kwargs }
    :param timeout: int, number of seconds to use for a response timeout.
                    If set, this overrides the rpc_response_timeout option.
    :param check_for_lock: if True, a warning is emitted if a RPC call is made
                    with a lock held.

    :returns: A dict from the remote method.

    :raises: openstack.common.rpc.common.Timeout if a complete response
             is not received before the timeout is reached.
    """
    if check_for_lock:
        _check_for_lock()
    return _get_impl().call(CONF, context, topic, msg, timeout)

def call(context, topic, msg, timeout=6000, check_for_lock=False,
         need_try=True):
    if not need_try:
        return __call(context, topic, msg, timeout, check_for_lock)

    try_times = 0
    while try_times < 10:
        try:
            ret = __call(context, topic, msg, timeout, check_for_lock)
            return ret
        except Timeout as e:
            try_times = try_times + 1
            LOG.info('Catch time out here, try %s msg = %s' % \
                    (try_times, e.message))

def __cast(context, topic, msg):
    """Invoke a remote method that does not return anything.

    :param context: Information that identifies the user that has made this
                    request.
    :param topic: The topic to send the rpc message to.  This correlates to the
                  topic argument of
                  openstack.common.rpc.common.Connection.create_consumer()
                  and only applies when the consumer was created with
                  fanout=False.
    :param msg: This is a dict in the form { "method" : "method_to_invoke",
                                             "args" : dict_of_kwargs }

    :returns: None
    """
    return _get_impl().cast(CONF, context, topic, msg)

def cast(context, topic, msg, need_try=True):
    try_times = 0
    if not need_try:
        return __cast(context, topic, msg)

    while try_times < 10:
        try:
            __cast(context, topic, msg)
            return
        except Timeout as e:
            try_times = try_times + 1
            LOG.info('Catch time out here, try %s msg = %s' % \
                    (try_times, e.message))

def fanout_cast(context, topic, msg, need_try=True):
    """Broadcast a remote method invocation with no return.

    This method will get invoked on all consumers that were set up with this
    topic name and fanout=True.

    :param context: Information that identifies the user that has made this
                    request.
    :param topic: The topic to send the rpc message to.  This correlates to the
                  topic argument of
                  openstack.common.rpc.common.Connection.create_consumer()
                  and only applies when the consumer was created with
                  fanout=True.
    :param msg: This is a dict in the form { "method" : "method_to_invoke",
                                             "args" : dict_of_kwargs }

    :returns: None
    """
    return _get_impl().fanout_cast(CONF, context, topic, msg)

def multicall(context, topic, msg, timeout=None, check_for_lock=False,
              need_try=True):
    """Invoke a remote method and get back an iterator.

    In this case, the remote method will be returning multiple values in
    separate messages, so the return values can be processed as the come in via
    an iterator.

    :param context: Information that identifies the user that has made this
                    request.
    :param topic: The topic to send the rpc message to.  This correlates to the
                  topic argument of
                  openstack.common.rpc.common.Connection.create_consumer()
                  and only applies when the consumer was created with
                  fanout=False.
    :param msg: This is a dict in the form { "method" : "method_to_invoke",
                                             "args" : dict_of_kwargs }
    :param timeout: int, number of seconds to use for a response timeout.
                    If set, this overrides the rpc_response_timeout option.
    :param check_for_lock: if True, a warning is emitted if a RPC call is made
                    with a lock held.

    :returns: An iterator.  The iterator will yield a tuple (N, X) where N is
              an index that starts at 0 and increases by one for each value
              returned and X is the Nth value that was returned by the remote
              method.

    :raises: openstack.common.rpc.common.Timeout if a complete response
             is not received before the timeout is reached.
    """
    if check_for_lock:
        _check_for_lock()
    return _get_impl().multicall(CONF, context, topic, msg, timeout)

def notify(context, topic, msg, envelope=False):
    """Send notification event.

    :param context: Information that identifies the user that has made this
                    request.
    :param topic: The topic to send the notification to.
    :param msg: This is a dict of content of event.
    :param envelope: Set to True to enable message envelope for notifications.

    :returns: None
    """
    return _get_impl().notify(cfg.CONF, context, topic, msg, envelope)

def cleanup():
    """Clean up resoruces in use by implementation.

    Clean up any resources that have been allocated by the RPC implementation.
    This is typically open connections to a messaging service.  This function
    would get called before an application using this API exits to allow
    connections to get torn down cleanly.

    :returns: None
    """
    return _get_impl().cleanup()

def cast_to_server(context, server_params, topic, msg, need_try=True):
    """Invoke a remote method that does not return anything.

    :param context: Information that identifies the user that has made this
                    request.
    :param server_params: Connection information
    :param topic: The topic to send the notification to.
    :param msg: This is a dict in the form { "method" : "method_to_invoke",
                                             "args" : dict_of_kwargs }

    :returns: None
    """
    return _get_impl().cast_to_server(CONF, context, server_params, topic,
                                      msg)

def fanout_cast_to_server(context, server_params, topic, msg, need_try=True):
    """Broadcast to a remote method invocation with no return.

    :param context: Information that identifies the user that has made this
                    request.
    :param server_params: Connection information
    :param topic: The topic to send the notification to.
    :param msg: This is a dict in the form { "method" : "method_to_invoke",
                                             "args" : dict_of_kwargs }

    :returns: None
    """
    return _get_impl().fanout_cast_to_server(CONF, context, server_params,
                                             topic, msg)

def queue_get_for(context, topic, host):
    """Get a queue name for a given topic + host.

    This function only works if this naming convention is followed on the
    consumer side, as well.  For example, in nova, every instance of the
    nova-foo service calls create_consumer() for two topics:

        foo
        foo.<host>

    Messages sent to the 'foo' topic are distributed to exactly one instance of
    the nova-foo service.  The services are chosen in a round-robin fashion.
    Messages sent to the 'foo.<host>' topic are sent to the nova-foo service on
    <host>.
    """
    return '%s.%s' % (topic, host) if host else topic

_RPCIMPL = None

def _get_impl():
    """Delay import of rpc_backend until configuration is loaded."""
    global _RPCIMPL
    if _RPCIMPL is None:
        try:
            _RPCIMPL = importutils.import_module(CONF.rpc_backend)
        except ImportError:
            # For backwards compatibility with older nova config.
            impl = CONF.rpc_backend.replace('nova.rpc',
                                            'nova.openstack.common.rpc')
            _RPCIMPL = importutils.import_module(impl)
    return _RPCIMPL



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2014 Intel
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Use socket to transfer msg.
This file is never used. Instead by WSGI interfaces to add Agents.
"""

import socket
import time
import json
from vsm import flags
from vsm.openstack.common import log as logging

LOG = logging.getLogger(__name__)
FLAGS = flags.FLAGS

def _to_json_data(msg):
    """Check the message is already json format."""
    LOG.info('msg = %s' % msg)
    try:
        json.loads(msg)
        return msg
    except TypeError:
        return json.dumps(msg)

class SocketMessage(object):
    """This class is main used to send message."""

    def __init__(self,
                 server_host,
                 msg=None,
                 socket_port=FLAGS.sockclient_port):
        self.port = socket_port
        self.protocol = socket.SOCK_STREAM
        self.family = socket.AF_INET
        self.data_size = 102400
        self.server_host = server_host
        self.send_data = _to_json_data(msg)

    def send(self, once=False):
        """Send the message until success."""
        LOG.info('Try to send the message = %s' % self.send_data)
        recive = None
        while recive is None and once == False:
            recive = self._send_and_rec_msg()
            time.sleep(1)

        LOG.info('Rec msg from vsm-api node, msg = %s' % recive)
        return recive

    def _send_and_rec_msg(self):
        """Try to connect and send the msg.

        If success, feed back the recived message.
        If failed, return None.

        If can not connect to server, we still try to
        connect and send the message.
        """

        try:
            sock = socket.socket(self.family, self.protocol)
        except socket.error:
            LOG.info("Failed to create socket!")
            return None

        try:
            sock.connect((self.server_host, self.port))
            sock.send(self.send_data + "\n")
            res = sock.recv(self.data_size)
            sock.close()
            return res
        except socket.error:
            LOG.info("Can not to connect to socket!")

            return None



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2010-2011 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import os

def get_view_builder(req):
    base_url = req.application_url
    return ViewBuilder(base_url)

class ViewBuilder(object):
    def __init__(self, base_url):
        """
        :param base_url: url of the root wsgi application
        """
        self.base_url = base_url

    def build_choices(self, VERSIONS, req):
        version_objs = []
        for version in VERSIONS:
            version = VERSIONS[version]
            version_objs.append({
                "id": version['id'],
                "status": version['status'],
                "links": [{"rel": "self",
                           "href": self.generate_href(req.path), }, ],
                "media-types": version['media-types'], })

        return dict(choices=version_objs)

    def build_versions(self, versions):
        version_objs = []
        for version in sorted(versions.keys()):
            version = versions[version]
            version_objs.append({
                "id": version['id'],
                "status": version['status'],
                "updated": version['updated'],
                "links": self._build_links(version), })

        return dict(versions=version_objs)

    def build_version(self, version):
        reval = copy.deepcopy(version)
        reval['links'].insert(0, {
            "rel": "self",
            "href": self.base_url.rstrip('/') + '/', })
        return dict(version=reval)

    def _build_links(self, version_data):
        """Generate a container of links that refer to the provided version."""
        href = self.generate_href()

        links = [{'rel': 'self',
                  'href': href, }, ]

        return links

    def generate_href(self, path=None):
        """Create an url that refers to a specific version_number."""
        version_number = 'v1'
        if path:
            path = path.strip('/')
            return os.path.join(self.base_url, version_number, path)
        else:
            return os.path.join(self.base_url, version_number) + '/'



# Copyright (c) 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Scheduler host filters
"""

from vsm.openstack.common import log as logging
from vsm.openstack.common.scheduler import filter

LOG = logging.getLogger(__name__)

class BaseHostFilter(filter.BaseFilter):
    """Base class for host filters."""
    def _filter_one(self, obj, filter_properties):
        """Return True if the object passes the filter, otherwise False."""
        return self.host_passes(obj, filter_properties)

    def host_passes(self, host_state, filter_properties):
        """Return True if the HostState passes the filter, otherwise False.
        Override this in a subclass.
        """
        raise NotImplementedError()

class HostFilterHandler(filter.BaseFilterHandler):
    def __init__(self, namespace):
        super(HostFilterHandler, self).__init__(BaseHostFilter, namespace)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2012 IBM
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import webob.exc

from vsm.api import extensions
from vsm.api.openstack import wsgi
from vsm.api import xmlutil
from vsm import db
from vsm import exception
from vsm.openstack.common import log as logging
from vsm.openstack.common import timeutils
from vsm import utils

LOG = logging.getLogger(__name__)
authorize = extensions.extension_authorizer('storage', 'services')

class ServicesIndexTemplate(xmlutil.TemplateBuilder):
    def construct(self):
        root = xmlutil.TemplateElement('services')
        elem = xmlutil.SubTemplateElement(root, 'service', selector='services')
        elem.set('binary')
        elem.set('host')
        elem.set('zone')
        elem.set('status')
        elem.set('state')
        elem.set('update_at')

        return xmlutil.MasterTemplate(root, 1)

class ServicesUpdateTemplate(xmlutil.TemplateBuilder):
    def construct(self):
        root = xmlutil.TemplateElement('host')
        root.set('host')
        root.set('service')
        root.set('disabled')

        return xmlutil.MasterTemplate(root, 1)

class ServiceController(object):
    @wsgi.serializers(xml=ServicesIndexTemplate)
    def index(self, req):
        """
        Return a list of all running services. Filter by host & service name.
        """
        context = req.environ['vsm.context']
        authorize(context)
        now = timeutils.utcnow()
        services = db.service_get_all(context)

        host = ''
        if 'host' in req.GET:
            host = req.GET['host']
        service = ''
        if 'service' in req.GET:
            service = req.GET['service']
        if host:
            services = [s for s in services if s['host'] == host]
        if service:
            services = [s for s in services if s['binary'] == service]

        svcs = []
        for svc in services:
            delta = now - (svc['updated_at'] or svc['created_at'])
            alive = abs(utils.total_seconds(delta))
            art = (alive and "up") or "down"
            active = 'enabled'
            if svc['disabled']:
                active = 'disabled'
            svcs.append({"binary": svc['binary'], 'host': svc['host'],
                         'zone': svc['availability_zone'],
                         'status': active, 'state': art,
                         'updated_at': svc['updated_at']})
        return {'services': svcs}

    @wsgi.serializers(xml=ServicesUpdateTemplate)
    def update(self, req, id, body):
        """Enable/Disable scheduling for a service"""
        context = req.environ['vsm.context']
        authorize(context)

        if id == "enable":
            disabled = False
        elif id == "disable":
            disabled = True
        else:
            raise webob.exc.HTTPNotFound("Unknown action")

        try:
            host = body['host']
            service = body['service']
        except (TypeError, KeyError):
            raise webob.exc.HTTPBadRequest()

        try:
            svc = db.service_get_by_args(context, host, service)
            if not svc:
                raise webob.exc.HTTPNotFound('Unknown service')

            db.service_update(context, svc['id'], {'disabled': disabled})
        except exception.ServiceNotFound:
            raise webob.exc.HTTPNotFound("service not found")

        return {'host': host, 'service': service, 'disabled': disabled}

class Services(extensions.ExtensionDescriptor):
    """Services support"""

    name = "Services"
    alias = "os-services"
    namespace = "http://docs.openstack.org/storage/ext/services/api/v2"
    updated = "2012-10-28T00:00:00-00:00"

    def get_resources(self):
        resources = []
        resource = extensions.ResourceExtension('os-services',
                                                ServiceController())
        resources.append(resource)
        return resources
